
class NotSparseError(AttributeError): pass


def raise_non_sparse_metric_or_loss_error():
    sparse_err = "mpunet 0.1.3 or higher requires integer targets" \
                 " as opposed to one-hot encoded targets. All metrics and " \
                 "loss functions should be named 'sparse_[org_name]' to " \
                 "reflect this change in accordance with the naming convention" \
                 " of TensorFlow.keras."
    raise NotSparseError(sparse_err)



import subprocess
import os
import pkgutil


class VersionController(object):
    def __init__(self, logger=None, package='mpunet'):
        from mpunet.logging.default_logger import ScreenLogger
        self.package_name = package
        self.package_loader = pkgutil.get_loader(package)
        self.logger = logger or ScreenLogger()
        self.git_path = os.path.split(os.path.split(self.package_loader.path)[0])[0]
        self._mem_path = None

    def log_version(self, logger=None):
        logger = logger or self.logger
        logger("{} version: {} ({}, {})".format(self.package_name,
                                                self.version,
                                                self.branch,
                                                self.current_commit))

    def __enter__(self):
        self._mem_path = os.getcwd()
        os.chdir(self.git_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._mem_path)
        self._mem_path = None

    def check_git(self):
        return bool(self.git_query("git status")) and \
               os.path.exists(self.git_path + "/.git")

    def git_query(self, string):
        with self:
            try:
                p = subprocess.Popen(string.split(),
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
                out, err = p.communicate()
            except FileNotFoundError:
                return None
            if not out:
                return None
            out = out.decode("utf-8").strip(" \n")
        return out

    @property
    def remote_url(self):
        return self.git_query("git config --get remote.origin.url")

    @property
    def version(self):
        module = self.package_loader.load_module()
        return module.__version__

    @property
    def current_commit(self):
        return self.git_query("git rev-parse --short HEAD")

    @property
    def latest_commit_in_branch(self, branch=None):
        branch = branch or self.branch
        url = self.remote_url
        commit = self.git_query("git ls-remote {} refs/heads/{}".format(
            url, branch
        ))
        if commit is None:
            raise OSError("Could not determine latest commit, did not find git.")
        return commit[:7]

    @property
    def branch(self):
        return self.git_query("git symbolic-ref --short HEAD")

    def set_commit(self, commit_id):
        self.git_query("git reset --hard {}".format(str(commit_id)[:7]))

    def set_branch(self, branch):
        self.git_query("git checkout {}".format(branch))

    def set_version(self, version):
        version = str(version).lower().strip(" v")
        self.set_branch("v{}".format(version))
        self.set_commit(self.latest_commit_in_branch)



"""
See
https://blog.keras.io/how-convolutional-neural-networks-see-the-world.html
"""

from tensorflow._api.v1.keras import backend as K
from PIL import Image as pil_image
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import os


def _get_figure(figsize, rows=1, cols=1):
    figsize = (figsize, figsize) if isinstance(figsize, int) else figsize
    fig, axes = plt.subplots(rows, cols, figsize=figsize)
    fig.subplots_adjust(hspace=0.005, wspace=0.005)
    if not isinstance(axes, (list, np.ndarray)):
        axes = [axes]
    if not isinstance(axes[0], (list, np.ndarray)):
        axes = [axes]
    for row in axes:
        for ax in row:
            ax.set_aspect("equal", adjustable="box")
            ax.axis("off")
    return fig, axes


def display_filter(filter, out_path=None, dpi=200, figsize=6, overwrite=False):
    fig, ax = _get_figure(figsize)
    ax[0].imshow(filter[0, :, :, 0], cmap="gray")
    fig.tight_layout()
    out_path = out_path or "filter.png"
    if not overwrite and os.path.exists(out_path):
        raise OSError("Out path {} already exists.".format(out_path))
    fig.savefig(out_path, dpi=dpi)
    plt.close(fig)


def display_filter_grid(filters, out_path=None, dpi=300, fig_width=8,
                        overwrite=False, layer_names=None):
    filters = np.array(filters)
    if filters.ndim == 3:
        filters = np.expand_dims(filters, 0)

    # Compute figure height and get figure with subplots
    rows, cols = filters.shape[0:2]
    ax_size = fig_width / cols
    fig_height = ax_size * rows
    fig, axes = _get_figure((fig_width, fig_height), rows=rows, cols=cols)

    # Plot the filters
    for layer_ind, (row, row_filters) in enumerate(zip(np.asarray(axes),
                                                       filters)):
        vmin, vmax = np.min(row_filters), np.max(row_filters)
        for i, (ax, filter) in enumerate(zip(row, row_filters)):
            ax.imshow(filter, cmap="gray", vmin=vmin, vmax=vmax)
            if i == 0:
                name = layer_names[layer_ind] if layer_names else layer_ind
                if len(name) > 10:
                    name = name[:10] + "\n" + name[10:]
                ax.annotate(s="{}".format(name),
                            xy=(-0.2, 0.5),
                            xycoords=ax.transAxes,
                            ha="center", va="center",
                            rotation=90, size=ax_size * 7)
    # Save image
    out_path = out_path or "filters.png"
    if not overwrite and os.path.exists(out_path):
        raise OSError("Out path {} already exists.".format(out_path))
    fig.savefig(out_path, dpi=dpi)
    plt.close(fig)


def deprocess_image(x):
    """utility function to convert a float array into a valid uint8 image.
    # Arguments
        x: A numpy-array representing the generated image.
    # Returns
        A processed numpy-array, which could be used in e.g. imshow.
    """
    # normalize tensor: center on 0., ensure std is 0.25
    x -= x.mean()
    x /= (x.std() + K.epsilon())
    x *= 0.25

    # clip to [0, 1]
    x += 0.5
    x = np.clip(x, 0, 1)

    # convert to RGB array
    x *= 255
    if K.image_data_format() == 'channels_first':
        x = x.transpose((1, 2, 0))
    x = np.clip(x, 0, 255).astype('uint8')
    return x


def process_image(x, former):
    """utility function to convert a valid uint8 image back into a float array.
       Reverses `deprocess_image`.
    # Arguments
        x: A numpy-array, which could be used in e.g. imshow.
        former: The former numpy-array.
                Need to determine the former mean and variance.
    # Returns
        A processed numpy-array representing the generated image.
    """
    if K.image_data_format() == 'channels_first':
        x = x.transpose((2, 0, 1))
    return (x / 255 - 0.5) * 4 * former.std() + former.mean()


def _run_optim(img, iterate, dim, step, steps):
    loss_value = 0.0
    for i in range(steps):
        s = "[*] Size {} - Step {}/{} - " \
            "loss {:.3f}".format(dim, i + 1, steps, loss_value)
        print(s, end="\r", flush=True)

        loss_value, grads_value = iterate([img])
        img += grads_value * step
        if loss_value <= K.epsilon():
            # some filters get stuck to 0, we can skip them
            break
    return img, loss_value


def visualize_filter(model,
                     layer_ind,
                     filter_index,
                     steps=20,
                     lr=1.0,
                     upscaling_steps=9,
                     upscaling_factor=1.2,
                     output_dim=(384, 384)):

    # Build loss function for the specified filter of the specified layer
    layer = model.layers[layer_ind]
    if not isinstance(layer, tf.keras.layers.Conv2D):
        raise ValueError("Currently supports only Conv2D layers, "
                         "got {}".format(type(layer)))
    loss = K.mean(K.abs(layer.output[..., filter_index]))
    input_image = model.inputs[0]
    grads = K.gradients(loss, input_image)[0]

    # normalization trick: we normalize the gradient
    grads /= (K.sqrt(K.mean(K.square(grads))) + 1e-5)

    # this function returns the loss and grads given the input picture
    iterate_func = K.function([input_image], [loss, grads])

    intermediate_dim = tuple(
        int(x / (upscaling_factor ** upscaling_steps)) for x in output_dim
    )
    img = np.random.randn(1, intermediate_dim[0], intermediate_dim[1], 1) * 0.1

    for up in reversed(range(upscaling_steps)):
        # we run gradient ascent for e.g. 20 steps
        img, loss_value = _run_optim(img, iterate_func, intermediate_dim,
                                     lr, steps)

        # Calculate up-scaled dimensions
        intermediate_dim = tuple(
            int(x / (upscaling_factor ** up)) for x in output_dim
        )
        # Upscale
        temp_img = deprocess_image(img[0])
        temp_img = np.stack((temp_img.squeeze(),)*3, axis=-1)
        temp_img = pil_image.fromarray(temp_img)
        temp_img = np.array(temp_img.resize(intermediate_dim,
                                            pil_image.BICUBIC))[..., 0:1]
        img = np.array([process_image(temp_img, img[0])])

    # Run one final time on full-size image
    img, loss_value = _run_optim(img, iterate_func, intermediate_dim,
                                 lr, steps)
    print("")
    return img.squeeze(), loss_value


if __name__ == "__main__":
    from mpunet.bin.predict import get_model, load_hparams
    from mpunet.utils import await_and_set_free_gpu

    p = "/home/jovyan/work/projects/knee_projects/" \
        "iwoai_challenge/splits/split_0"
    # Get model
    await_and_set_free_gpu(1)
    hparams = load_hparams(p)
    model = get_model(p, 1, hparams["build"])[0]

    # SETTINGS
    layer_range = range(0, len(model.layers))
    filter_range = range(0, 200)
    n_highest = 9
    steps = 30
    lr = 0.5
    upscaling_steps = 19

    per_layer_filters = []
    layer_names = []
    for layer_ind in layer_range:
        layer = model.layers[layer_ind]
        print("\n[*] Layer {}".format(layer_ind))
        if not isinstance(layer, tf.keras.layers.Conv2D):
            print("--- skipping (not Conv2D)")
            continue
        filters, losses = [], []
        for filter_ind in filter_range:
            try:
                filter, loss = visualize_filter(model=model,
                                                layer_ind=layer_ind,
                                                steps=steps,
                                                lr=lr,
                                                upscaling_steps=upscaling_steps,
                                                filter_index=filter_ind)
            except ValueError as e:
                break
            filters.append(filter), losses.append(loss)

        # Save filters with highest losses
        inds = np.argsort(losses)[::-1][:n_highest]
        per_layer_filters.append([filters[i] for i in inds])
        layer_names.append(layer.name)

        display_filter_grid(per_layer_filters,
                            overwrite=True,
                            layer_names=layer_names,
                            fig_width=10,
                            dpi=1200)



import numpy as np
from .base_sequence import BaseSequence


class RandomDataFeeder(BaseSequence):
    def __init__(self, X, y, batch_size):
        super().__init__()
        self.X = X
        self.y = y
        self.bs = batch_size
        self.inds = np.arange(len(self.X))

    def __len__(self):
        n_samples = len(self.X)
        return int(np.ceil(n_samples / self.bs))

    def __getitem__(self, item):
        inds = np.random.choice(self.inds, size=self.bs, replace=False)
        return self.X[inds], self.y[inds]



import numpy as np
from mpunet.interpolation.regular_grid_interpolator import RegularGridInterpolator

from mpunet.preprocessing import reshape_add_axis
from mpunet.interpolation.linalg import mgrid_to_points, points_to_mgrid
from mpunet.interpolation.sample_grid import get_voxel_grid, get_voxel_axes_real_space, get_voxel_grid_real_space
from mpunet.sequences.utils import get_sequence


def predict_single(image, model, hparams, verbose=1):
    """
    A generic prediction function that sets up a ImagePairLoader object for the
    given image, prepares the image and predicts.

    Note that this function should only be used for convinience in scripts that
    work on single images at a time anyway, as batch-preparing the entire
    ImagePairLoader object prior to prediction is faster.

    NOTE: Only works with iso_live intrp modes at this time
    """
    mode = hparams["fit"]["intrp_style"].lower()
    assert mode in ("iso_live", "iso_live_3d")

    # Create a ImagePairLoader with only the given file
    from mpunet.image import ImagePairLoader
    image_pair_loader = ImagePairLoader(predict_mode=True,
                                        initialize_empty=True,
                                        no_log=bool(verbose))
    image_pair_loader.add_image(image)

    # Set scaler and bg values
    image_pair_loader.set_scaler_and_bg_values(
        bg_value=hparams.get_from_anywhere('bg_value'),
        scaler=hparams.get_from_anywhere('scaler'),
        compute_now=False
    )

    if mode == "iso_live":
        # Init LazyQueue and get its sequencer
        seq = get_sequence(data_queue=image_pair_loader,
                           views=np.load(hparams.project_path + "/views.npz")["arr_0"],
                           is_validation=True,
                           **hparams["fit"], **hparams["build"])

        with seq.image_pair_queue.get_image_by_id(image.identifier) as image:
            # Get voxel grid in real space
            voxel_grid_real_space = get_voxel_grid_real_space(image)

            # Prepare tensor to store combined prediction
            d = image.image.shape
            predicted = np.empty(shape=(len(seq.views), d[0], d[1], d[2], seq.n_classes),
                                 dtype=np.float32)
            print("Predicting on brain hyper-volume of shape:", predicted.shape)

            for n_view, v in enumerate(seq.views):
                print("\nView %i/%i: %s" % (n_view+1, len(seq.views), v))
                # Sample the volume along the view
                X, y, grid, inv_basis = seq.get_view_from(image, v,
                                                          n_planes="same+20")

                # Predict on volume using model
                pred = predict_volume(model, X, axis=2)

                # Map the real space coordiante predictions to nearest
                # real space coordinates defined on voxel grid
                predicted[n_view] = map_real_space_pred(pred, grid, inv_basis,
                                                        voxel_grid_real_space,
                                                        method="nearest")
    else:
        # Init LazyQueue and get its sequencer
        seq = get_sequence(data_queue=image_pair_loader,
                           is_validation=True,
                           **hparams["fit"], **hparams["build"])
        predicted = pred_3D_iso(model=model, sequence=seq,
                                image=image, extra_boxes="3x",
                                min_coverage=None)

    return predicted


def predict_volume(model, X, batch_size=8, axis=0):
    """
    TODO
    """
    print("Predicting...")
    # Move axis to front
    X = np.moveaxis(X, source=axis, destination=0)
    pred = model.predict(X, batch_size=batch_size, verbose=1)
    return np.moveaxis(pred, source=0, destination=axis)


def map_real_space_pred(pred, grid, inv_basis, voxel_grid_real_space, method="nearest"):
    """
    TODO
    """
    print("Mapping to real coordinate space...")

    # Prepare fill value vector, we set this to 1.0 background
    fill = np.zeros(shape=pred.shape[-1], dtype=np.float32)
    fill[0] = 1.0

    # Initialize interpolator object
    intrp = RegularGridInterpolator(grid, pred, fill_value=fill,
                                    bounds_error=False, method=method)

    points = inv_basis.dot(mgrid_to_points(voxel_grid_real_space).T).T
    transformed_grid = points_to_mgrid(points, voxel_grid_real_space[0].shape)

    # Prepare mapped pred volume
    mapped = np.empty(transformed_grid[0].shape + (pred.shape[-1],),
                      dtype=pred.dtype)

    # Prepare interpolation function
    def _do(xs, ys, zs, index):
        return intrp((xs, ys, zs)), index

    # Prepare thread pool of 10 workers
    from concurrent.futures import ThreadPoolExecutor
    pool = ThreadPoolExecutor(max_workers=7)

    # Perform interpolation async.
    inds = np.arange(transformed_grid.shape[1])
    result = pool.map(_do, transformed_grid[0], transformed_grid[1],
                      transformed_grid[2], inds)

    i = 1
    for map, ind in result:
        # Print status
        print("  %i/%i" % (i, inds[-1]+1), end="\r", flush=True)
        i += 1

        # Map the interpolation results into the volume
        mapped[ind] = map

    print("")
    pool.shutdown()
    return mapped


def predict_3D_patches_binary(model, patches, image_id, N_extra=0, logger=None):
    """
    TODO
    """
    # Get box dim and image dim
    d = patches.dim
    i1, i2, i3 = patches.im_dim

    # Prepare reconstruction volume. Predictions will be summed in this volume.
    recon = np.zeros(shape=(i1, i2, i3, 2), dtype=np.uint32)

    # Predict on base patches + N extra randomly
    # sampled patches from the volume
    for patch, (i, k, v), status in patches.get_patches_from(image_id, N_extra):
        # Log the status of the generator
        print(status, end="\r", flush=True)

        # Predict on patch
        pred = model.predict_on_batch(reshape_add_axis(patch, im_dims=3)).numpy().squeeze()
        mask = pred > 0.5

        # Add prediction to reconstructed volume
        recon[i:i+d, k:k+d, v:v+d, 0] += ~mask
        recon[i:i+d, k:k+d, v:v+d, 1] += mask
    print("")

    total = np.sum(recon, axis=-1)
    return (recon[..., 1] > (0.20 * total)).astype(np.uint8)


def predict_3D_patches(model, patches, image, N_extra=0, logger=None):
    """
    TODO
    """
    # Get box dim and image dim
    d = patches.dim
    i1, i2, i3 = image.shape[:3]

    # Prepare reconstruction volume. Predictions will be summed in this volume.
    recon = np.zeros(shape=(i1, i2, i3, model.n_classes), dtype=np.float32)

    # Predict on base patches + N extra randomly
    # sampled patches from the volume
    for patch, (i, k, v), status in patches.get_patches_from(image, N_extra):
        # Log the status of the generator
        print(status, end="\r", flush=True)

        # Predict on patch
        pred = model.predict_on_batch(reshape_add_axis(patch, im_dims=3)).numpy()

        # Add prediction to reconstructed volume
        recon[i:i+d, k:k+d, v:v+d] += pred.squeeze()
    print("")

    # Normalize
    recon /= np.sum(recon, axis=-1, keepdims=True)

    return recon


def pred_3D_iso(model, sequence, image, extra_boxes, min_coverage=None):
    """
    TODO
    """
    total_extra_boxes = extra_boxes

    # Get reference to the image
    n_classes = sequence.n_classes
    pred_shape = tuple(image.shape[:3]) + (n_classes,)
    vox_shape = tuple(image.shape[:3]) + (3,)

    # Prepare interpolator object
    vox_grid = get_voxel_grid(image, as_points=False)

    # Get voxel regular grid centered in real space
    g_all, basis, _ = get_voxel_axes_real_space(image.image, image.affine,
                                                return_basis=True)
    g_all = list(g_all)

    # Flip axes? Must be strictly increasing
    flip = np.sign(np.diagonal(basis)) == -1
    for i, (g, f) in enumerate(zip(g_all, flip)):
        if f:
            g_all[i] = np.flip(g, 0)
            vox_grid = np.flip(vox_grid, i+1)
    vox_points = mgrid_to_points(vox_grid).reshape(vox_shape).astype(np.float32)

    # Setup interpolator - takes a point in the scanner space and returns
    # the nearest voxel coordinate
    intrp = RegularGridInterpolator(tuple(g_all), vox_points,
                                    method="nearest", bounds_error=False,
                                    fill_value=np.nan, dtype=np.float32)

    # Prepare prediction volume
    pred_vol = np.zeros(shape=pred_shape, dtype=np.float32)

    # Predict on base patches first
    base_patches = sequence.get_base_patches_from(image, return_y=False)

    # Sample boxes and predict --> sum into pred_vol
    is_covered, base_reached, extra_reached, N_base, N_extra = not min_coverage, False, False, 0, 0

    while not is_covered or not base_reached or not extra_reached:
        try:
            im, rgrid, _, _, total_base = next(base_patches)
            N_base += 1

            if isinstance(total_extra_boxes, str):
                # Number specified in string format '2x', '2.5x' etc. as a
                # multiplier of number of base patches
                total_extra_boxes = int(float(total_extra_boxes.split("x")[0]) * total_base)

        except StopIteration:
            p = sequence.get_N_random_patches_from(image, 1, return_y=False)
            im, rgrid, _, _ = next(p)
            N_extra += 1

        # Predict on the box
        pred = model.predict_on_batch(np.expand_dims(im, 0))[0].numpy()

        # Apply rotation if needed
        rgrid = image.interpolator.apply_rotation(rgrid)

        # Interpolate to nearest vox grid positions
        vox_inds = intrp(tuple(rgrid)).reshape(-1, 3)

        # Flatten and mask results
        mask = np.logical_not(np.all(np.isnan(vox_inds), axis=-1))
        vox_inds = [i for i in vox_inds[mask].astype(np.int).T]

        # Add to volume
        pred_vol[tuple(vox_inds)] += pred.reshape(-1, n_classes)[mask]

        # Check coverage fraction
        if min_coverage:
            covered = np.logical_not(np.all(np.isclose(pred_vol, 0), axis=-1))
            coverage = np.sum(covered) / np.prod(pred_vol.shape[:3])
            cov_string = "%.3f/%.3f" % coverage, min_coverage
            is_covered = coverage >= min_coverage
        else:
            cov_string = "[Not calculated]"

        print("   N base patches: %i/%i --- N extra patches %i/%i --- "
              "Coverage: %s" % (
                N_base, total_base, N_extra, total_extra_boxes, cov_string),
              end="\r", flush=True)

        # Check convergence
        base_reached = N_base >= total_base
        extra_reached = N_extra >= total_extra_boxes
    print("")

    # Return prediction volume - OBS not normalized
    return pred_vol



from mpunet.interpolation.sample_grid import get_real_image_size, get_pix_dim
from mpunet.utils import highlighted
from mpunet.logging import ScreenLogger
import nibabel as nib
import numpy as np


def _audit_classes(nii_lab_paths, logger):
    logger("Auditing number of target classes. This may take "
           "a while as data must be read from disk."
           "\n-- Note: avoid this by manually setting the "
           "n_classes attribute in train_hparams.yaml.")
    # Select up to 50 random images and find the unique classes
    lab_paths = np.random.choice(nii_lab_paths,
                                 min(50, len(nii_lab_paths)),
                                 replace=False)
    classes = []
    for l in lab_paths:
        classes.extend(np.unique(nib.load(l).get_data()))
    classes = np.unique(classes)
    n_classes = classes.shape[0]

    # Make sure the classes start from 0 and step continuously by 1
    c_min, c_max = np.min(classes), np.max(classes)
    if c_min != 0:
        raise ValueError("Invalid class audit - Class integers should"
                         " start from 0, found %i (classes found: %s)"
                         % (c_min, classes))
    if n_classes != max(classes) + 1:
        raise ValueError("Invalid class audit - Found %i classes, but"
                         " expected %i, as the largest class value"
                         " found was %i. Classes found: %s"
                         % (n_classes, c_max+1, c_max, classes))
    return n_classes


class Auditor(object):
    """
    Parses all .nii/.nii.gz images of a specified folder and proposes
    heuristically determined interpolation parameters for models working in
    isotropic scanner space coordinates.

    If label paths are specified, also audits the number of target classes
    for this segmentation task by sampling up to 50 images and noting the
    number of unique classes across them.

    Suggested parameters are stored for both 2D and 3D models on this object
    The selected parameters can be written to a mpunet YAMLHParams
    object which in turn may write them to the train_hparams.yaml file on disk.

    The heuristic covers 3 parameters:
        1) The sample dimension
           - The number of pixels to sample in each dim.
        2) The real space span
           - The number of mm along each axis that should define the sample
             space around the image (real space) center within which image may
             be sampled.
        3) Real box dimension (3D only)
           - The number of mm the sampled 3D box spans along all 3 axes

    In addition, the auditor stores the number of channels in the images and
    estimated the total memory needed to store all images in memory.

    See paper for a description of how the heuristic define those parameters

    ------
    NOTE: This class allows for fully autonomous use of the 2D mpunet
    and 3D UNet models when used with the mpunet.bin scrips.

    NOTE: The heuristic is not guaranteed to be optimal for all problems.
    """
    def __init__(self, nii_paths, nii_lab_paths=None, logger=None,
                 min_dim_2d=128, max_dim_2d=512, dim_3d=64, span_percentile=75,
                 res_percentile=25, hparams=None):
        """
        Args:
            nii_paths: A list of paths pointing to typically training and val
                       .nii/.nii.gz images to audit
            nii_lab_paths: Optional paths pointing to .nii/.nii.gz label images
                           from which target class number is inferred
            logger: A mpunet logger object
            min_dim_2d: Minimum pixel dimension to use
            max_dim_2d: Maximum pixel dimension to use (usually GPU limited)
            dim_3d: Pixel dimensionality of the 3D model
            span_percentile: The real space span param will be set close to the
                             'span_percentile' percentile computed across all
                             spans recorded across images and axes.
            res_percentile: The sampled resolution will be set close to the
                            'span_percentile' percentile computed across all
                            voxel resolutions recorded across images and axes.
        """
        self.nii_paths = nii_paths
        self.nii_lab_paths = nii_lab_paths
        self.logger = logger or ScreenLogger()

        # Fetch basic information on the images
        self.hparms = hparams
        self.info = self.audit()

        """ Set some attributes used for image sampling """
        assert np.all(np.array(self.info["n_channels"]) == self.info["n_channels"][0])
        self.n_channels = int(self.info["n_channels"][0])

        # Number of classes
        self.n_classes = self.info["n_classes"]

        # 2D
        real_space_span = np.percentile(self.info["real_sizes"], span_percentile)
        sample_res = np.percentile(self.info["pixdims"], res_percentile)
        self.sample_dim_2D, self.real_space_span_2D = self.heurestic_sample_dim(real_space_span,
                                                                                sample_res,
                                                                                min_dim_2d, max_dim_2d)

        # 3D
        self.sample_dim_3D = dim_3d
        self.real_space_span_3D = real_space_span
        self.real_box_span = dim_3d * sample_res

        # Total memory (including channels)
        self.total_memory_bytes = sum(self.info["memory_bytes"])
        self.total_memory_gib = self.total_memory_bytes/np.power(1024, 3)

        # Set hparams pattern
        # These patterns map a Auditor attribute to the sub-field and name
        # under this field in which the value should be stored in the
        # train_hparams.yaml file
        # Currently, these are specified broadly for 2D and 3D models
        # TODO: Integrate this with the logic of the mpunet.models
        # TODO: __init__.py file that already sets preprep functions for each
        # TODO: model type.
        self.patterns = {
            "2d": {
                "real_space_span_2D": (["fit"], ["real_space_span"]),
                "sample_dim_2D": (["build"], ["dim"]),
                "n_channels": (["build"], ["n_channels"]),
                "n_classes": (["build"], ["n_classes"])
            },
            "3d": {
                "real_space_span_3D": (["fit"], ["real_space_span"]),
                "sample_dim_3D": (["build"], ["dim"]),
                "real_box_span": (["fit"], ["real_box_dim"]),
                "n_channels": (["build"], ["n_channels"]),
                "n_classes": (["build"], ["n_classes"])
            },
            "multi_task_2d": {
                "real_space_span_2D": (["task_specifics"], ["real_space_span"]),
                "sample_dim_2D": (["task_specifics"], ["dim"]),
                "n_channels": (["task_specifics"], ["n_channels"]),
                "n_classes": (["task_specifics"], ["n_classes"])
            }
        }

        # Write to log
        self.log()

    def log(self):
        self.logger(highlighted("\nAudit for %i images" % len(self.nii_paths)))
        self.logger("Total memory GiB:  %.3f" % self.total_memory_gib)
        if self.n_classes is not None:
            self.logger("Number of classes: %i" % self.n_classes)
        self.logger("\n2D:\n"
                    "Real space span:   %.3f\n"
                    "Sample dim:        %.3f" % (self.real_space_span_2D,
                                                 self.sample_dim_2D))
        self.logger("\n3D:\n"
                    "Sample dim:        %i\n"
                    "Real space span:   %.3f\n"
                    "Box span:          %.3f" % (self.sample_dim_3D,
                                                 self.real_space_span_3D,
                                                 self.real_box_span))

    def fill(self, hparams, model_type):
        """
        Add and write attributes stored in this Auditor object to the
        YAMLHParams object and train_hparams.yaml file according to the
        patterns self.pattern_2d and self.pattern_3d (see init)

        Only attributes not already manually specified by the user will be
        changed. See YAMLHParams.set_value().

        Args:
            hparams:     mpunet YAMLHParams object
            model_type:  A string representing the model type and thus which
                         pattern to apply. Must be either "2d", "3d" (upper case tolerated)
        """
        model_type = model_type.lower()
        pattern = self.patterns.get(model_type)
        if pattern is None:
            raise ValueError("Unknown model type: '%s'" % model_type)

        for attr in pattern:
            subdirs, names = pattern[attr]
            value = getattr(self, attr)

            for s, n in zip(subdirs, names):
                hparams.set_value(subdir=s, name=n, value=value)
        hparams.save_current()

    def heurestic_sample_dim(self, real_space_span, res, _min, _max):
        valid = np.array([i for i in range(_min, _max+1) if (i*0.5**4).is_integer()])
        sample_dim = real_space_span / res
        nearest_valid = valid[np.abs(valid - sample_dim).argmin()]

        if nearest_valid < (sample_dim * 0.90):
            # Reduce real space span a bit to increase resolution
            pref = nearest_valid * res
            real_space_span = max(int(real_space_span * 0.70), pref)

        return nearest_valid, real_space_span

    def audit(self):
        shapes = []
        channels = []
        real_sizes = []
        pixdims = []
        memory = []

        for im_path in self.nii_paths:
            # Load the nii file without loading image data
            im = nib.load(im_path)

            # Get image voxel shape
            shape = im.shape
            shapes.append(shape[:3])

            try:
                c = shape[3]
            except IndexError:
                c = 1
            channels.append(c)

            # Get image real shape
            real_sizes.append(get_real_image_size(im))

            # Get pixel dims
            pixdims.append(get_pix_dim(im))

            # Calculate memory in bytes to store image
            memory.append(im.get_data_dtype().itemsize * np.prod(shape))

        n_classes = None
        if self.nii_lab_paths is not None:
            # Labels exists, thus we need the n_classes attribute
            if self.hparms is not None:
                # Attempt to get it from a potentially specified hparams obj
                n_classes = self.hparms.get_from_anywhere("n_classes")
            if n_classes is None:
                # If still none, infer it
                n_classes = _audit_classes(self.nii_lab_paths,
                                           self.logger)

        info = {
            "shapes": shapes,
            "real_sizes": real_sizes,
            "pixdims": pixdims,
            "memory_bytes": memory,
            "n_channels": channels,
            "n_classes": n_classes,
        }
        return info



from .base_sequence import BaseSequence
from .patch_sequence_3d import PatchSequence3D
from .sliding_patch_sequence_3d import SlidingPatchSequence3D
from .isotrophic_live_view_sequence_2d import IsotrophicLiveViewSequence2D
from .isotrophic_live_view_sequence_3d import IsotrophicLiveViewSequence3D
from .multi_task_sequence import MultiTaskSequence
from .simple_sequences import RandomDataFeeder



import argparse
from mpunet.tests import get_test_packages


def get_parser():
    parser = argparse.ArgumentParser(description="Run all or a set of tests.")
    parser.add_argument("--tests", type=str, default="all",
                        help="Select a package of tests to run. "
                             "Must be one of: (all, {}). "
                             "Defaults to "
                             "all.".format(", ".join(get_test_packages())))
    return parser


def entry_func(args=None):

    # Get parser
    parser = get_parser().parse_args(args)
    tests_to_run = parser.tests.lower()

    from mpunet.tests import run_tests
    try:
        run_tests(tests_to_run)
    except ValueError as e:
        print(str(e))


if __name__ == "__main__":
    entry_func()



from .logger import Logger
from .default_logger import ScreenLogger
from .log_results import init_result_dicts, save_all, init_result_dict_3D, \
                         save_all_3D, load_result_dicts



import tensorflow as tf
from tensorflow.python.keras.losses import LossFunctionWrapper


def _to_tensor(x, dtype):
    """Convert the input `x` to a tensor of type `dtype`.

    OBS: Code implemented by Tensorflow

    # Arguments
        x: An object to be converted (numpy array, list, tensors).
        dtype: The destination type.

    # Returns
        A tensor.
    """
    x = tf.convert_to_tensor(x)
    if x.dtype != dtype:
        x = tf.cast(x, dtype)
    return x


def _get_shapes_and_one_hot(y_true, y_pred):
    shape = y_pred.get_shape()
    n_classes = shape[-1]
    # Squeeze dim -1 if it is == 1, otherwise leave it
    dims = tf.cond(tf.equal(y_true.shape[-1] or -1, 1), lambda: tf.shape(y_true)[:-1], lambda: tf.shape(y_true))
    y_true = tf.reshape(y_true, dims)
    y_true = tf.one_hot(tf.cast(y_true, tf.uint8), depth=n_classes)
    return y_true, shape, n_classes


def sparse_jaccard_distance_loss(y_true, y_pred, smooth=1):
    """
    Jaccard = (|X & Y|)/ (|X|+ |Y| - |X & Y|)
            = sum(|A*B|)/(sum(|A|)+sum(|B|)-sum(|A*B|))

    The jaccard distance loss is usefull for unbalanced datasets. This has been
    shifted so it converges on 0 and is smoothed to avoid exploding or disapearing
    gradient.

    Approximates the class-wise jaccard distance computed per-batch element
    across spatial image dimensions. Returns the 1 - mean(per_class_distance)
    for each batch element.

    :param y_true:
    :param y_pred:
    :param smooth:
    :return:

    Ref: https://en.wikipedia.org/wiki/Jaccard_index

    @url: https://gist.github.com/wassname/f1452b748efcbeb4cb9b1d059dce6f96
    @author: wassname
    """
    y_true, shape, n_classes = _get_shapes_and_one_hot(y_true, y_pred)
    reduction_dims = range(len(shape))[1:-1]

    intersection = tf.reduce_sum(y_true * y_pred, axis=reduction_dims)
    sum_ = tf.reduce_sum(y_true + y_pred, axis=reduction_dims)
    jac = (intersection + smooth) / (sum_ - intersection + smooth)
    return 1.0 - tf.reduce_mean(jac, axis=-1, keepdims=True)


class SparseJaccardDistanceLoss(LossFunctionWrapper):
    """ tf reduction wrapper for sparse_jaccard_distance_loss """
    def __init__(self,
                 reduction,
                 smooth=1,
                 name='sparse_jaccard_distance_loss',
                 **kwargs):
        super(SparseJaccardDistanceLoss, self).__init__(
            sparse_jaccard_distance_loss,
            name=name,
            reduction=reduction,
            smooth=smooth
        )


def sparse_dice_loss(y_true, y_pred, smooth=1):
    """
    Approximates the class-wise dice coefficient computed per-batch element
    across spatial image dimensions. Returns the 1 - mean(per_class_dice) for
    each batch element.

    :param y_true:
    :param y_pred:
    :param smooth:
    :return:
    """
    y_true, shape, n_classes = _get_shapes_and_one_hot(y_true, y_pred)
    reduction_dims = range(len(shape))[1:-1]

    intersection = tf.reduce_sum(y_true * y_pred, axis=reduction_dims)
    union = tf.reduce_sum(y_true + y_pred, axis=reduction_dims)
    dice = (2 * intersection + smooth) / (union + smooth)
    return 1.0 - tf.reduce_mean(dice, axis=-1, keepdims=True)


class SparseDiceLoss(LossFunctionWrapper):
    """ tf reduction wrapper for sparse_dice_loss """
    def __init__(self,
                 reduction,
                 smooth=1,
                 name='sparse_dice_loss',
                 **kwargs):
        super(SparseDiceLoss, self).__init__(
            sparse_dice_loss,
            name=name,
            reduction=reduction,
            smooth=smooth
        )


def sparse_exponential_logarithmic_loss(y_true, y_pred, gamma_dice,
                                        gamma_cross, weight_dice,
                                        weight_cross):
    """
    TODO

    :param y_true:
    :param y_pred:
    :param smooth:
    :return:
    """
    y_true, shape, n_classes = _get_shapes_and_one_hot(y_true, y_pred)
    reduction_dims = range(len(shape))[1:-1]

    # Clip for numerical stability
    _epsilon = _to_tensor(10e-8, y_pred.dtype.base_dtype)
    y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)

    # Compute exp log dice
    intersect = 2 * tf.reduce_sum(y_true * y_pred, axis=reduction_dims) + 1
    union = tf.reduce_sum(y_true + y_pred, axis=reduction_dims) + 1
    exp_log_dice = tf.math.pow(-tf.math.log(intersect/union), gamma_dice)
    mean_exp_log_dice = tf.reduce_mean(exp_log_dice, axis=-1, keepdims=True)

    # Compute exp cross entropy
    entropy = tf.reduce_sum(y_true * -tf.math.log(y_pred), axis=-1, keepdims=True)
    exp_entropy = tf.reduce_mean(tf.math.pow(entropy, gamma_cross), axis=reduction_dims)

    # Compute output
    res = weight_dice*mean_exp_log_dice + weight_cross*exp_entropy
    return res


class SparseExponentialLogarithmicLoss(LossFunctionWrapper):
    """
    https://link.springer.com/content/pdf/10.1007%2F978-3-030-00931-1_70.pdf
    """
    def __init__(self, reduction, gamma_dice=0.3, gamma_cross=0.3,
                 weight_dice=1, weight_cross=1,
                 name="sparse_exponential_logarithmic_loss"):
        super(SparseExponentialLogarithmicLoss, self).__init__(
            sparse_exponential_logarithmic_loss,
            name=name,
            reduction=reduction,
            gamma_dice=gamma_dice,
            gamma_cross=gamma_cross,
            weight_dice=weight_dice,
            weight_cross=weight_cross
        )


def sparse_focal_loss(y_true, y_pred, gamma, class_weights):
    """
    TODO

    :param y_true:
    :param y_pred:
    :param smooth:
    :return:
    """
    y_true, shape, n_classes = _get_shapes_and_one_hot(y_true, y_pred)
    reduction_dims = range(len(shape))[1:-1]

    # Clip for numerical stability
    _epsilon = _to_tensor(10e-8, y_pred.dtype.base_dtype)
    y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)

    if class_weights is None:
        class_weights = [1] * n_classes

    # Compute the focal loss
    entropy = tf.math.log(y_pred)
    modulator = tf.math.pow((1 - y_pred), gamma)
    loss = -tf.reduce_sum(class_weights * y_true * modulator * entropy, axis=-1, keepdims=True)
    return tf.reduce_mean(loss, axis=reduction_dims)


class SparseFocalLoss(LossFunctionWrapper):
    """
    https://arxiv.org/pdf/1708.02002.pdf
    """
    def __init__(self, reduction, gamma=2,
                 class_weights=None, name="sparse_focal_loss"):
        super(SparseFocalLoss, self).__init__(
            sparse_focal_loss,
            name=name,
            reduction=reduction,
            gamma=gamma,
            class_weights=class_weights
        )


def sparse_generalized_dice_loss(y_true, y_pred, type_weight):
    """
    Function to calculate the Generalised Dice Loss defined in
        Sudre, C. et. al. (2017) Generalised Dice overlap as a deep learning
        loss function for highly unbalanced segmentations. DLMIA 2017
    """
    y_true, shape, n_classes = _get_shapes_and_one_hot(y_true, y_pred)
    reduction_dims = range(len(shape))[1:-1]

    ref_vol = tf.reduce_sum(y_true, axis=reduction_dims)
    intersect = tf.reduce_sum(y_true * y_pred, axis=reduction_dims)
    seg_vol = tf.reduce_sum(y_pred, axis=reduction_dims)

    if type_weight.lower() == 'square':
        weights = tf.math.reciprocal(tf.math.square(ref_vol))
    elif type_weight.lower() == 'simple':
        weights = tf.math.reciprocal(ref_vol)
    elif type_weight.lower() == 'uniform':
        weights = tf.ones_like(ref_vol)
    else:
        raise ValueError("The variable type_weight \"{}\""
                         "is not defined.".format(type_weight))

    # Make array of new weight in which infinite values are replaced by
    # ones.
    new_weights = tf.where(tf.math.is_inf(weights),
                           tf.zeros_like(weights),
                           weights)

    # Set final weights as either original weights or highest observed
    # non-infinite weight
    weights = tf.where(tf.math.is_inf(weights), tf.ones_like(weights) *
                       tf.reduce_max(new_weights), weights)

    # calculate generalized dice score
    eps = 1e-6
    numerator = 2 * tf.multiply(weights, intersect)
    denom = tf.multiply(weights, seg_vol + ref_vol) + eps
    generalised_dice_score = numerator / denom
    return 1 - tf.reduce_mean(generalised_dice_score, axis=-1, keepdims=True)


class SparseGeneralizedDiceLoss(LossFunctionWrapper):
    """
    Based on implementation in NiftyNet at:

    http://niftynet.readthedocs.io/en/dev/_modules/niftynet/layer/
    loss_segmentation.html#generalised_dice_loss

    Class based to allow passing of parameters to the function at construction
    time in keras.
    """
    def __init__(self, reduction, type_weight="Square",
                 name='sparse_generalized_dice_loss'):
        super(SparseGeneralizedDiceLoss, self).__init__(
            sparse_generalized_dice_loss,
            name=name,
            reduction=reduction,
            type_weight=type_weight
        )


# Aliases
SparseExpLogDice = SparseExponentialLogarithmicLoss



import os
import numpy as np
import random
import argparse

from mpunet.utils import create_folders
from glob import glob


def get_parser():
    parser = argparse.ArgumentParser(description="Prepare a data folder for a"
                                                 "CV experiment setup.")
    parser.add_argument("--data_dir", type=str,
                        help="Path to data directory")
    parser.add_argument("--CV", type=int, default=5,
                        help="Number of splits (default=5)")
    parser.add_argument("--out_dir", type=str, default="views",
                        help="Directory to store CV subfolders "
                             "(default=views")
    parser.add_argument("--im_sub_dir", type=str, default="images",
                        help="Subfolder under 'data_dir' in which image are "
                             "stored (default=images)")
    parser.add_argument("--lab_sub_dir", type=str, default="labels",
                        help="Subfolder under 'data_dir' in which labels are "
                             "stored (default=labels)")
    parser.add_argument("--copy", action="store_true",
                        help="Copy files to CV-subfolders instead of "
                             "symlinking (not recommended)")
    parser.add_argument("--file_list", action="store_true",
                        help="Create text files with paths pointing to the "
                             "images at the image and labels subdirs under "
                             "each split instead of symlink/copying. This is"
                             " usefull on systems were symlink is not "
                             "supported, but the dataset size is too large to"
                             " store in copies. NOTE: Only one of --copy and "
                             "--file_list flags must be set.")
    parser.add_argument("--file_regex", type=str, default="*.nii*",
                        help="Regex used to select files from the image "
                             "and labels subdirs. (default='*.nii*')")
    parser.add_argument("--validation_fraction", type=float, default=0.20,
                        help="Fraction of OVERALL data size used for "
                             "validation in each split. In a 5-CV setting with "
                             "N=100 and val_frac=0.20, each split will have "
                             "N_train=60, N_val=20 and N_test=20 images")
    parser.add_argument("--test_fraction", type=float, default=0.20,
                        help="Fraction of data size used for test if CV=1.")
    parser.add_argument("--common_prefix_length", type=int, required=False, default=0)
    return parser


def assert_dir_structure(data_dir, im_dir, lab_dir, out_dir):
    for _dir in (data_dir, im_dir, lab_dir):
        if not os.path.exists(_dir):
            raise OSError("Invalid data directory '%s'. Does not exist." % data_dir)
    if os.path.exists(out_dir):
        raise OSError("Output directory at '%s' already exists." % out_dir)


def create_view_folders(out_dir, n_splits):
    if not os.path.exists(out_dir):
        print("Creating directory at %s" % out_dir)
        os.makedirs(out_dir)

    if n_splits > 1:
        for i in range(n_splits):
            split_dir = os.path.join(out_dir, "split_%i" % i)
            print("Creating directory at %s" % split_dir)
            os.mkdir(split_dir)


def pair_by_names(images, common_prefix_length):
    if common_prefix_length == 0:
        return images
    from collections import defaultdict
    names = [os.path.split(i)[-1][:common_prefix_length] for i in images]
    inds = defaultdict(list)
    for i, item in enumerate(names):
        inds[item].append(i)
    pairs = inds.values()
    return [tuple(np.array(images)[i]) for i in pairs]


def add_images(images, im_folder_path, label_folder_path, im_dir, lab_dir,
               link_func=os.symlink):
    for image in images:
        if not isinstance(image, (list, tuple, np.ndarray)):
            image = (image,)
        for im in image:
            # Get file name
            file_name = os.path.split(im)[-1]

            # Get label path (OBS: filenames must match!)
            lab = im.replace(im_dir, lab_dir)

            if not os.path.exists(lab):
                raise OSError("No label file found at '%s'. OBS: image and "
                              "label files must have exactly the same name. "
                              "Images should be located at '%s' and labels at"
                              " '%s'" % (lab, im_folder_path, label_folder_path))

            # Get relative paths
            rel_image = os.path.relpath(im, im_folder_path)
            rel_label = os.path.relpath(lab, label_folder_path)

            # Symlink or copy
            link_func(rel_image, im_folder_path + "/%s" % file_name)
            link_func(rel_label, label_folder_path + "/%s" % file_name)


def _add_to_file_list_fallback(rel_image_path, image_path,
                               fname="LIST_OF_FILES.txt"):

    """
    On some system synlinks are not supported, if --files_list flag is set,
    uses this function to add each absolute file path to a list at the final
    subfolder that is supposed to store images and label links or actual files

    At run-time, these files must be loaded by reading in the path from these
    files instead.
    """
    # Get folder where list of files should be stored
    folder = os.path.split(image_path)[0]

    # Get absolute path to image
    # We change dir to get the correct abs path from the relative
    os.chdir(folder)
    abs_file_path = os.path.abspath(rel_image_path)

    # Get path to the list of files
    list_file_path = os.path.join(folder, fname)

    with open(list_file_path, "a") as out_f:
        out_f.write(abs_file_path + "\n")


def entry_func(args=None):

    # Get parser
    parser = vars(get_parser().parse_args(args))

    # Get arguments
    data_dir = os.path.abspath(parser["data_dir"])
    n_splits = int(parser["CV"])
    if n_splits > 1:
        out_dir = os.path.join(data_dir, parser["out_dir"], "%i_CV" % n_splits)
    else:
        out_dir = os.path.join(data_dir, parser["out_dir"], "fixed_split")
    im_dir = os.path.join(data_dir, parser["im_sub_dir"])
    lab_dir = os.path.join(data_dir, parser["lab_sub_dir"])

    copy = parser["copy"]
    file_list = parser["file_list"]
    regex = parser["file_regex"]
    val_frac = parser["validation_fraction"]
    test_frac = parser["test_fraction"]
    common_prefix_length = parser["common_prefix_length"]

    if n_splits == 1 and not test_frac:
        raise ValueError("Must specify --test_fraction with --CV=1.")
    if copy and file_list:
        raise ValueError("Only one of --copy and --file_list "
                         "flags must be set.")

    # Assert suitable folders
    assert_dir_structure(data_dir, im_dir, lab_dir, out_dir)

    # Create sub-folders
    create_view_folders(out_dir, n_splits)

    # Get images and pair by subject identifier if common_prefix_length > 0
    images = glob(os.path.join(im_dir, regex))
    images = pair_by_names(images, common_prefix_length)
    print("-----")
    print("Found {} images".format(len(images)))

    # Get validation size
    N_total = len(images)
    if n_splits > 1:
        N_test = N_total // n_splits
    else:
        N_test = int(np.ceil(N_total * test_frac))
    N_val = int(np.ceil(N_total * val_frac))
    if N_val + N_test >= N_total:
        raise ValueError("Too large validation_fraction - "
                         "No training samples left!")
    N_train = N_total - N_test - N_val
    print("Total images:".ljust(40), N_total)
    print("Train images pr. split:".ljust(40), N_train)
    print("Validation images pr. split:".ljust(40), N_val)
    print("Test images pr. split:".ljust(40), N_test)

    # Shuffle and split the images into CV parts
    random.shuffle(images)
    splits = np.array_split(images, n_splits)

    # Symlink / copy files
    for i, split in enumerate(splits):
        print("  Split %i/%i" % (i+1, n_splits), end="\r", flush=True)

        # Set root path to split folder
        if n_splits > 1:
            split_path = os.path.join(out_dir, "split_%i" % i)
        else:
            split_path = out_dir
            # Here we kind of hacky force the following code to work with CV=1
            # Define a test set and overwrite the current split (which stores
            # add the data, as splits was never split with n_splits=1
            split = splits[0][:N_test]

            # Overwrite the splits variable to a length 2 array with the
            # remaining data which will be used as val+train. The loop still
            # refers to the old split and thus will only execute once
            splits = [split, splits[0][N_test:]]

        # Define train, val and test sub-dirs
        train_path = os.path.join(split_path, "train")
        train_im_path = os.path.join(train_path, parser["im_sub_dir"])
        train_label_path = os.path.join(train_path, parser["lab_sub_dir"])
        if N_val:
            val_path = os.path.join(split_path, "val")
            val_im_path = os.path.join(val_path, parser["im_sub_dir"])
            val_label_path = os.path.join(val_path, parser["lab_sub_dir"])
        else:
            val_path, val_im_path, val_label_path = (None,) * 3
        test_path = os.path.join(split_path, "test")
        test_im_path = os.path.join(test_path, parser["im_sub_dir"])
        test_label_path = os.path.join(test_path, parser["lab_sub_dir"])

        # Create folders if not existing
        create_folders([train_path, val_path, train_im_path, train_label_path,
                        val_im_path, val_label_path, test_path, test_im_path,
                        test_label_path])

        # Copy or symlink?
        if copy:
            from shutil import copyfile
            move_func = copyfile
        elif file_list:
            move_func = _add_to_file_list_fallback
        else:
            move_func = os.symlink

        # Add test data to test folder
        add_images(split, test_im_path, test_label_path, im_dir, lab_dir, move_func)

        # Join remaining splits into train+val
        remaining = [x for ind, x in enumerate(splits) if ind != i]
        remaining = [item for sublist in remaining for item in sublist]

        # Extract validation data from the remaining
        random.shuffle(remaining)
        validation = remaining[:N_val]
        training = remaining[N_val:]

        # Add
        if validation:
            add_images(validation, val_im_path, val_label_path, im_dir, lab_dir, move_func)
        add_images(training, train_im_path, train_label_path, im_dir, lab_dir, move_func)


if __name__ == "__main__":
    entry_func()



from mpunet.sequences.isotrophic_live_view_sequence import IsotrophicLiveViewSequence
from mpunet.interpolation.sample_grid import sample_box, sample_box_at
from mpunet.interpolation.linalg import mgrid_to_points
import numpy as np


class IsotrophicLiveViewSequence3D(IsotrophicLiveViewSequence):
    def __init__(self, image_pair_queue, real_box_dim, no_log=False, **kwargs):
        super().__init__(image_pair_queue, **kwargs)

        self.real_box_dim = real_box_dim
        self.batch_shape = (self.batch_size, self.sample_dim, self.sample_dim,
                            self.sample_dim, self.n_classes)

        if not no_log:
            self.log()

    def log(self):
        self.logger("Using sample dim:            %s" % self.sample_dim)
        self.logger("Using box real dim:          %s" % self.real_box_dim)
        self.logger("Using real space sample res: %s" % (self.real_box_dim/
                                                         self.sample_dim))
        self.logger("N fg slices:                 %s" % self.n_fg_slices)
        self.logger("Batch size:                  %s" % self.batch_size)
        self.logger("Force all FG:                %s" % self.force_all_fg)

    @staticmethod
    def _intrp_and_norm(image, grid, intrp_lab):
        # Interpolate
        im = image.interpolator.intrp_image(grid)

        # Normalize
        im = image.scaler.transform(im)

        lab = None
        if intrp_lab:
            lab = image.interpolator.intrp_labels(grid)

        return im, lab

    def get_base_patches_from(self, image, return_y=False):
        real_dims = image.real_shape

        # Calculate positions
        sample_space = np.asarray([max(i, self.real_box_dim) for i in real_dims])
        d = (sample_space - self.real_box_dim)
        min_cov = [np.ceil(sample_space[i]/self.real_box_dim).astype(np.int) for i in range(3)]
        ds = [np.linspace(0, d[i], min_cov[i]) - sample_space[i]/2 for i in range(3)]

        # Get placement coordinate points
        placements = mgrid_to_points(np.meshgrid(*tuple(ds)))

        for p in placements:
            grid, axes, inv_mat = sample_box_at(real_placement=p,
                                                sample_dim=self.sample_dim,
                                                real_box_dim=self.real_box_dim,
                                                noise_sd=0.0,
                                                test_mode=True)

            im, lab = self._intrp_and_norm(image, grid, return_y)

            if return_y:
                yield im, lab, grid, axes, inv_mat, len(placements)
            else:
                yield im, grid, axes, inv_mat, len(placements)

    def get_N_random_patches_from(self, image, N, return_y=False):
        if N > 0:
            # Sample N patches from X
            for i in range(N):
                # Get grid and interpolate
                grid, axes, inv_mat = sample_box(sample_dim=self.sample_dim,
                                                 real_box_dim=self.real_box_dim,
                                                 real_dims=image.real_shape,
                                                 noise_sd=self.noise_sd,
                                                 test_mode=True)

                im, lab = self._intrp_and_norm(image, grid, return_y)

                if return_y:
                    yield im, lab, grid, axes, inv_mat
                else:
                    yield im, grid, axes, inv_mat
        else:
            return []

    def _get_valid_box_from(self, image, max_tries, has_fg_vec, has_fg_count, cur_bs):
        """
        TODO
        """
        tries = 0
        while tries < max_tries:
            # Sample a batch from the image
            tries += 1

            # Get grid and interpolate
            mgrid = sample_box(sample_dim=self.sample_dim,
                               real_box_dim=self.real_box_dim,
                               real_dims=image.real_shape,
                               noise_sd=self.noise_sd)

            # Get interpolated labels
            lab = image.interpolator.intrp_labels(mgrid)
            valid_lab, fg_change = self.validate_lab(lab, has_fg_count, cur_bs)

            if self.force_all_fg and tries < max_tries:
                valid, has_fg_vec = self.validate_lab_vec(lab, has_fg_vec, cur_bs)
                if not valid:
                    continue

            if valid_lab or tries == max_tries:
                # Get interpolated image
                im = image.interpolator.intrp_image(mgrid)
                im_bg_val = image.interpolator.bg_value
                if tries == max_tries or self.is_valid_im(im, im_bg_val):
                    # Accept box, update foreground counter
                    has_fg_vec += fg_change
                    return im, lab, has_fg_count

    def __getitem__(self, idx):
        """
        Used by keras.fit_generator to fetch mini-batches during training
        """
        # If multiprocessing, set unique seed for this particular process
        self.seed()

        # Store how many slices has fg so far
        has_fg_count = 0
        has_fg_vec = np.zeros_like(self.fg_classes)

        # Interpolate on a random index for each sample image to generate batch
        batch_x, batch_y, batch_w = [], [], []

        # Get a random image
        max_tries = self.batch_size * 10

        scalers = []
        bg_values = []
        for _ in range(self.batch_size):
            with self.image_pair_queue.get_random_image() as image:
                im, lab, has_fg_count = self._get_valid_box_from(
                    image=image,
                    max_tries=max_tries,
                    has_fg_vec=has_fg_vec,
                    has_fg_count=has_fg_count,
                    cur_bs=len(batch_y)
                )

                # Save scaler to normalize image later (after potential aug)
                scalers.append(image.scaler)

                # Save bg value if needed in potential augmenters
                bg_values.append(image.interpolator.bg_value)

                # Add to batches
                batch_x.append(im)
                batch_y.append(lab)
                batch_w.append(image.sample_weight)

        # Normalize images
        batch_x = self.scale(batch_x, scalers)

        # Apply augmentation if specified
        batch_x, batch_y, batch_w = self.augment(batch_x, batch_y,
                                                 batch_w, bg_values)

        # Reshape, one-hot encode etc.
        batch_x, batch_y, batch_w = self.prepare_batches(batch_x,
                                                         batch_y,
                                                         batch_w)

        assert len(batch_x) == self.batch_size
        return batch_x, batch_y, batch_w



import numpy as np


def one_hot_encode_y(y, n_classes=135):
    y = np.asarray(y)
    if n_classes == 1:
        return reshape_add_axis(y, len(y.shape)-1)
    else:
        from keras.utils import to_categorical
        shape = y.shape
        y = to_categorical(y, num_classes=n_classes).astype(np.uint8)
        y = y.reshape(shape + (n_classes,))
    return y


def reshape_add_axis(X, im_dims=2, n_channels=1):
    X = np.asarray(X)
    if X.shape[-1] != n_channels:
        # Reshape
        X = X.reshape(X.shape + (n_channels,))
    if len(X.shape) == im_dims+1:
        X = X.reshape((1,) + X.shape)
    return X



import numpy as np
import pandas as pd

import tensorflow as tf
from tensorflow.keras.callbacks import Callback
from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from collections import defaultdict
from mpunet.logging import ScreenLogger
from mpunet.evaluate.metrics import dice_all
from mpunet.utils import highlighted, ensure_list_or_tuple


class Validation(Callback):
    """
    Validation computation callback.
    Samples a number of validation batches from a mpunet.sequence object
    and computes for all tasks:
        - Batch-wise validation loss + metrics as specified in model.compile
        - Epoch-wise pr-class and average precision
        - Epoch-wise pr-class and average recall
        - Epoch-wise pr-class and average dice coefficients
    ... and adds all results to the log dict
    Note: The purpose of this callback over the default tf.keras evaluation
    mechanism is to calculate certain metrics over the entire epoch of data as
    opposed to averaged batch-wise computations.
    Also, it supports multi-dataset/multi-task evaluation
    """
    def __init__(self, val_sequence, steps, logger=None, verbose=True,
                 ignore_class_zero=True):
        """
        Args:
            val_sequence: A mpunet.sequence object from which validation
                          batches can be sampled via its __getitem__ method.
            steps:        Numer of batches to sample from val_sequences in each
                          validation epoch
            logger:       An instance of a MultiPlanar Logger that prints to screen
                          and/or file
            verbose:      Print progress to screen - OBS does not use Logger
            ignore_class_zero: TODO
        """
        super().__init__()
        self.logger = logger or ScreenLogger()
        self.data = val_sequence
        self.steps = steps
        self.verbose = verbose
        self.ignore_bg = ignore_class_zero
        self.print_round = 3
        self.log_round = 4
        self._supports_tf_logs = True  # ensures correct logs passed from tf.keras

        self.n_classes = self.data.n_classes
        if isinstance(self.n_classes, int):
            self.task_names = [""]
            self.n_classes = [self.n_classes]
        else:
            self.task_names = self.data.task_names

    @staticmethod
    def _compute_dice(tp, rel, sel):
        """
        TODO

        :param tp:
        :param rel:
        :param sel:
        :return:
        """
        # Get data masks (to avoid div. by zero warnings)
        # We set precision, recall, dice to 0 in for those particular cls.
        sel_mask = sel > 0
        rel_mask = rel > 0

        # prepare arrays
        precisions = np.zeros(shape=tp.shape, dtype=np.float32)
        recalls = np.zeros_like(precisions)
        dices = np.zeros_like(precisions)

        # Compute precisions, recalls
        precisions[sel_mask] = tp[sel_mask] / sel[sel_mask]
        recalls[rel_mask] = tp[rel_mask] / rel[rel_mask]

        # Compute dice
        intrs = (2 * precisions * recalls)
        union = (precisions + recalls)
        dice_mask = union > 0
        dices[dice_mask] = intrs[dice_mask] / union[dice_mask]

        return precisions, recalls, dices

    @staticmethod
    def _count_cm_elements_from_queue(queue,
                                      steps,
                                      TPs,
                                      relevant,
                                      selected,
                                      task_names,
                                      n_classes_list,
                                      lock):
        """
        TODO

        :param queue:
        :param steps:
        :param TPs:
        :param relevant:
        :param selected:
        :param task_names:
        :param n_classes_list:
        :param lock:
        :return:
        """
        for _ in range(steps):
            # Get prediction and true labels from prediction queue
            pred, true = queue.get(block=True)
            for p, y, task_name, n_classes in zip(pred, true, task_names,
                                                  n_classes_list):
                # Argmax and CM elements
                if not isinstance(p, np.ndarray):
                    p = p.numpy()
                p = p.argmax(-1).ravel()
                y = y.ravel()

                # Compute relevant CM elements
                # We select the number following the largest class integer
                # when y != pred, then bincount and remove the added dummy
                # class
                tps = np.bincount(np.where(y == p, y, n_classes),
                                  minlength=n_classes+1)[:-1]
                rel = np.bincount(y, minlength=n_classes)
                sel = np.bincount(p, minlength=n_classes)

                # Update counts on shared lists
                lock.acquire()
                TPs[task_name] += tps.astype(np.uint64)
                relevant[task_name] += rel.astype(np.uint64)
                selected[task_name] += sel.astype(np.uint64)
                lock.release()
            queue.task_done()

    def evalaute(self):
        """
        TODO

        :return:
        """
        # Get tensors to run and their names
        if hasattr(self.model, "loss_functions"):
            metrics = self.model.loss_functions + self.model.metrics
        else:
            metrics = self.model.metrics
        metrics_names = self.model.metrics_names
        self.model.reset_metrics()
        assert len(metrics_names) == len(metrics)

        # Prepare dictionaries for storing pr. task metric results
        TPs, relevant, selected, batch_wise_metrics = {}, {}, {}, {}
        for task_name, n_classes in zip(self.task_names, self.n_classes):
            TPs[task_name] = np.zeros(shape=(n_classes,), dtype=np.uint64)
            relevant[task_name] = np.zeros(shape=(n_classes,), dtype=np.uint64)
            selected[task_name] = np.zeros(shape=(n_classes,), dtype=np.uint64)
            batch_wise_metrics[task_name] = defaultdict(list)

        # Prepare queue and thread for computing counts
        from queue import Queue
        from threading import Thread
        count_queue = Queue(maxsize=self.steps)
        count_thread = Thread(target=self._count_cm_elements_from_queue,
                              daemon=True,
                              args=[count_queue, self.steps, TPs, relevant,
                                    selected, self.task_names, self.n_classes,
                                    Lock()])
        count_thread.start()

        # Fetch validation batches from the generator(s)
        pool = ThreadPoolExecutor(max_workers=3)
        batches = pool.map(self.data.__getitem__, np.arange(self.steps))

        # Predict on all
        self.logger("")
        for i, (X, y, _) in enumerate(batches):
            if self.verbose:
                print("   Validation: %i/%i" % (i+1, self.steps),
                      end="\r", flush=True)
            X = ensure_list_or_tuple(X)
            y = ensure_list_or_tuple(y)

            # Predict and put values in the queue for counting
            pred = self.model.predict_on_batch(X)
            pred = ensure_list_or_tuple(pred)
            count_queue.put([pred, y])

            for p_task, y_task, task in zip(pred, y, self.task_names):
                # Run all metrics
                for metric, name in zip(metrics, metrics_names):
                    m = tf.reduce_mean(metric(y_task, p_task))
                    batch_wise_metrics[task][name].append(m.numpy())
        pool.shutdown(wait=True)

        # Compute the mean over batch-wise metrics
        mean_batch_wise_metrics = {}
        for task in self.task_names:
            mean_batch_wise_metrics[task] = {}
            for metric in metrics_names:
                ms = batch_wise_metrics[task][metric]
                mean_batch_wise_metrics[task][metric] = np.mean(ms)
        self.model.reset_metrics()
        self.logger("")

        # Terminate count thread
        print("Waiting for counting queue to terminate...\n")
        count_thread.join()
        count_queue.join()

        # Compute per-class metrics (dice+precision+recall)
        class_wise_metrics = {}
        for task in self.task_names:
            precisions, recalls, dices = self._compute_dice(tp=TPs[task],
                                                            sel=relevant[task],
                                                            rel=selected[task])
            if self.ignore_bg:
                precisions[0] = np.nan
                recalls[0] = np.nan
                dices[0] = np.nan
            class_wise_metrics[task] = {
                "dice": dices,
                "recall": recalls,
                "precision": precisions
            }
        return class_wise_metrics, mean_batch_wise_metrics

    def _print_val_results(self,
                           class_wise_metrics,
                           batch_wise_metrics,
                           epoch,
                           task_name,
                           classes):
        """
        TODO

        :param class_wise_metrics:
        :param batch_wise_metrics:
        :param epoch:
        :param task_name:
        :param classes:
        :return:
        """
        # We add them to a pd dataframe just for the pretty print output
        index = ["mean"] + ["cls %i" % i for i in classes]
        columns = list(batch_wise_metrics.keys()) + list(class_wise_metrics.keys())
        df = pd.DataFrame(data={c: [np.nan] * len(index) for c in columns},
                          index=index)

        # Fill the df with metrics
        for m_name, value in batch_wise_metrics.items():
            df.loc['mean', m_name] = value
        for m_name, values in class_wise_metrics.items():
            values = [np.nanmean(values)] + list(values)
            df.loc[:, m_name] = values

        # Print the df to screen
        s = "Validation Results for epoch %i" % epoch
        self.logger(highlighted((("[%s]" % task_name) if task_name else "") + s))
        print_string = df.round(self.print_round).T.to_string()
        self.logger(print_string.replace("NaN", "  -") + "\n")

    def on_epoch_end(self, epoch, logs={}):
        # Predict and get CM
        class_wise_metrics, mean_batch_wise_metrics = self.evalaute()
        for n_classes, name in zip(self.n_classes, self.task_names):
            classes = np.arange(n_classes)
            n = (name + "_") if len(self.task_names) > 1 else ""

            # Add batch-wise metrics to log
            for m_name, value in mean_batch_wise_metrics[name].items():
                logs[f"{n}val_{m_name}"] = value.round(self.log_round)
            # Add mean of class-wise metrics to log
            for m_name, values in class_wise_metrics[name].items():
                logs[f"{n}val_{m_name}"] = np.nanmean(values)

            if self.verbose:
                self._print_val_results(class_wise_metrics=class_wise_metrics[name],
                                        batch_wise_metrics=mean_batch_wise_metrics[name],
                                        epoch=epoch,
                                        task_name=name,
                                        classes=classes)

        if len(self.task_names) > 1:
            # Print cross-dataset mean values
            if self.verbose:
                self.logger(highlighted(f"[ALL DATASETS] Means Across Classes"
                                        f" for Epoch {epoch}"))
            fetch = ("val_dice", "val_precision", "val_recall")
            m_fetch = tuple(["val_" + s for s in self.model.metrics_names])
            to_print = {}
            for f in fetch + m_fetch:
                scores = [logs["%s_%s" % (name, f)] for name in self.task_names]
                res = np.mean(scores)
                logs[f] = res.round(self.log_round)  # Add to log file
                to_print[f.split("_")[-1]] = list(scores) + [res]
            if self.verbose:
                df = pd.DataFrame(to_print)
                df.index = self.task_names + ["mean"]
                self.logger(df.round(self.print_round))
            self.logger("")


class ValDiceScores(Callback):
    """
    Similar to Validation, but working on an array of data instead of
    internally sampling from a validation sequence generator.

    On epoch end computes the mean dice coefficient and adds to following log
    entry:
    logs["val_dice"] = mean_dice
    """
    def __init__(self, validation_data, n_classes, batch_size=8, logger=None):
        """
        Args:
            validation_data: A tuple (X, y) of two ndarrays of validation data
                             and corresponding labels.
                             Any shape accepted by the model.
                             Labels must be integer targets (not one-hot)
            n_classes:       Number of classes, including background
            batch_size:      Batch size used for prediction
            logger:          An instance of a MultiPlanar Logger that prints to screen
                             and/or file
        """
        super().__init__()
        self.logger = logger or ScreenLogger()
        self.X_val, self.y_val = validation_data
        self.n_classes = n_classes
        self.batch_size = batch_size
        self.scores = []

    def eval(self):
        pred = self.model.predict(self.X_val, self.batch_size, verbose=1)
        dices = dice_all(self.y_val,
                         pred.argmax(-1),
                         n_classes=self.n_classes,
                         ignore_zero=True)
        return dices

    def on_epoch_end(self, epoch, logs={}):
        scores = self.eval()
        mean_dice = scores.mean()
        s = "Mean dice for epoch %d: %.4f\nPr. class: %s" % (epoch,
                                                             mean_dice,
                                                             scores)
        self.logger(highlighted(s))
        self.scores.append(mean_dice)

        # Add to log
        logs["val_dice"] = mean_dice



from mpunet.sequences.isotrophic_live_view_sequence import IsotrophicLiveViewSequence
from mpunet.interpolation.sample_grid import sample_plane_at, get_bounding_sphere_real_radius
import numpy as np


class IsotrophicLiveViewSequence2D(IsotrophicLiveViewSequence):
    def __init__(self, image_pair_queue, views, no_log=False, **kwargs):

        super().__init__(image_pair_queue, **kwargs)
        self.views = views
        self.batch_shape = (self.batch_size, self.sample_dim, self.sample_dim,
                            self.n_classes)

        if not no_log:
            self.log()

    def log(self):
        self.logger("\nIs validation:               %s" % self.is_validation)
        self.logger("Using real space span:       %s" % self.real_space_span)
        self.logger("Using sample dim:            %s" % self.sample_dim)
        self.logger("Using real space sample res: %s" % (self.real_space_span/
                                                         self.sample_dim))
        self.logger("N fg slices:                 %s" % self.n_fg_slices)
        self.logger("Batch size:                  %s" % self.batch_size)
        self.logger("Force all FG:                %s" % self.force_all_fg)
        self.logger("Noise SD:                    %s" % self.noise_sd)
        self.logger("Augmenters:                  %s" % self.list_of_augmenters)

    def get_view_from(self, image, view, n_planes):
        """
        TODO

        :param image:
        :param view:
        :param n_planes:
        :return:
        """
        # Prepare sample plane arguments
        kwargs = {
            "norm_vector": view,
            "sample_dim": self.sample_dim,
            "real_space_span": self.real_space_span,
            "noise_sd": 0.,
            "test_mode": True
        }

        sample_res = self.real_space_span/(self.sample_dim-1)
        if n_planes == "by_radius":
            # Get sample sphere radius
            bounds = get_bounding_sphere_real_radius(image)
            n_planes = int(2 * bounds / sample_res)
        else:
            extra = 0
            if n_planes == "same":
                n_planes = self.sample_dim
            elif isinstance(n_planes, str) and n_planes[:5] == "same+":
                extra = int(n_planes.split("+")[-1])
                n_planes = self.sample_dim + extra
            bounds = (self.real_space_span+(extra*sample_res))/2

        # Define offsets
        offsets = np.linspace(-bounds, bounds, n_planes)
        self.logger("Sampling %i planes from "
                    "offset %.3f to %.3f..." % (n_planes, offsets[0],
                                                offsets[-1]))

        # Prepare results arrays
        shape = (self.sample_dim, self.sample_dim, n_planes)
        Xs = np.empty(shape + (image.n_channels,), dtype=image.image.dtype)
        if not image.predict_mode:
            ys = np.empty(shape, dtype=image.labels.dtype)
        else:
            ys = None

        # Prepare thread pool
        from concurrent.futures import ThreadPoolExecutor
        pool = ThreadPoolExecutor(max_workers=7)

        def _do(offset, ind):
            im, lab, real_axis, inv_basis = self.sample_at(offset,
                                                           image.interpolator,
                                                           image.scaler,
                                                           kwargs)
            return im, lab, real_axis, inv_basis, ind

        # Perform interpolation
        inds = np.arange(offsets.shape[0])
        result = pool.map(_do, offsets, inds)

        i = 1
        for im, lab, real_axis, inv_basis, ind in result:
            print("   %i/%i" % (i, len(offsets+1)), end="\r", flush=True)
            i += 1

            # Add planes to volumes
            Xs[..., ind, :] = im
            if not image.predict_mode:
                ys[..., ind] = lab

        print('')
        return Xs, ys, (real_axis, real_axis, offsets), inv_basis

    def sample_at(self, offset, interpolator, scaler, kwargs):
        """
        TODO
        """
        # Get plane mgrid
        grid, real_axis, inv_basis = sample_plane_at(offset_from_center=offset,
                                                     **kwargs)

        # Interpolate at grid points
        im, lab = interpolator(grid)

        # Normalize
        im = scaler.transform(im)

        return im, lab, real_axis, inv_basis

    def _get_valid_slice_from(self, image, max_tries, has_fg_vec, has_fg_count, cur_bs):
        """
        TODO
        """
        tries = 0
        while tries < max_tries:
            # Sample a batch from the image
            tries += 1

            # Randomly sample a slice from a random image and random view
            view = self.views[np.random.randint(0, len(self.views), 1)[0]]

            # Get sample sphere radius
            sphere_r_real = self.real_space_span // 2

            # Sample a position on the axis
            rd = np.random.uniform(-sphere_r_real, sphere_r_real, 1)[0]

            # Get grid and interpolate
            mgrid = sample_plane_at(view,
                                    sample_dim=self.sample_dim,
                                    real_space_span=self.real_space_span,
                                    offset_from_center=rd,
                                    noise_sd=self.noise_sd,
                                    test_mode=False)

            # Get interpolated labels
            lab = image.interpolator.intrp_labels(mgrid)

            if self.force_all_fg and tries < max_tries:
                valid, has_fg_vec = self.validate_lab_vec(lab, has_fg_vec, cur_bs)
                if not valid:
                    continue

            valid_lab, fg_change = self.validate_lab(lab, has_fg_count, cur_bs)
            if valid_lab or tries == max_tries:
                # Get interpolated image
                im = image.interpolator.intrp_image(mgrid)
                im_bg_val = image.interpolator.bg_value
                if tries == max_tries or self.is_valid_im(im, im_bg_val):
                    # Accept slice + update foreground counter
                    has_fg_count += fg_change
                    return im, lab, has_fg_count

    def __getitem__(self, idx):
        """
        TODO
        """
        # If multiprocessing, set unique seed for this particular process
        self.seed()

        # Store how many slices has fg so far
        has_fg_count = 0
        has_fg_vec = np.zeros_like(self.fg_classes)

        # Interpolate on a random index for each sample image to generate batch
        batch_x, batch_y, batch_w = [], [], []

        # Maximum number of sampling trails
        max_tries = 10

        scalers = []
        bg_values = []
        for _ in range(self.batch_size):
            with self.image_pair_queue.get_random_image() as image:
                im, lab, has_fg_count = self._get_valid_slice_from(
                    image=image,
                    max_tries=max_tries,
                    has_fg_vec=has_fg_vec,
                    has_fg_count=has_fg_count,
                    cur_bs=len(batch_y)
                )

                # Save scaler to normalize image later (after potential aug)
                scalers.append(image.scaler)

                # Save bg value if needed in potential augmenters
                bg_values.append(image.interpolator.bg_value)

                # Add to batches
                batch_x.append(im)
                batch_y.append(lab)
                batch_w.append(image.sample_weight)

        # Normalize images
        batch_x = self.scale(batch_x, scalers)

        # Apply augmentation if specified
        batch_x, batch_y, batch_w = self.augment(batch_x, batch_y,
                                                 batch_w, bg_values)

        # Reshape, one-hot encode etc.
        batch_x, batch_y, batch_w = self.prepare_batches(batch_x,
                                                         batch_y,
                                                         batch_w)

        assert len(batch_x) == self.batch_size
        return batch_x, batch_y, batch_w



from .fuse_and_predict import predict_volume, \
                              predict_3D_patches, predict_3D_patches_binary, \
                              map_real_space_pred, pred_3D_iso, predict_single
from .fusion_training import stack_collections, predict_and_map



import glob
import os
import sys
from argparse import ArgumentParser
from shutil import copy

from mpunet.train import YAMLHParams
from mpunet.utils import get_best_model, create_folders


def get_parser():
    parser = ArgumentParser(description='Branch a trained multi-task model '
                                        'into separate models for each task.')

    parser.add_argument("--project_folder", type=str,
                        default=os.path.abspath("./"),
                        help="The project folder (defaults to current dir)")
    parser.add_argument("--copy_weights", action="store_true",
                        help="Copy the weights file instead of sym-linking.")
    parser.add_argument("--weights_file", type=str, default="",
                        help="(Optional) Path to the weights file. Will be"
                             " inferred automatically if not specified.")
    parser.add_argument("--out_dir", type=str, default="branches",
                        help="Sub-folder to store the branched projects ("
                             "defaults to 'branches')")

    return parser


def branch(task_name, out_dir, task_hparams, task_hparams_file,
           shared_hparams, weights, copy_weights, views_file):
    create_folders(out_dir)

    # Set the task fields to only the current task under 'build'
    shared_hparams.set_value("build", "task_names",
                             value='["%s"]' % task_name, overwrite=True)
    shared_hparams.delete_group("tasks")  # No longer necessary

    # Create a map defining where each task specific parameter should go
    mapping = {
        "task_specifics/n_classes": "build/n_classes",
        "task_specifics/n_channels": "build/n_channels",
        "task_specifics/dim": ["build/dim", "fit/dim"],
        "task_specifics/out_activation": "build/out_activation",
        "task_specifics/real_space_span": "fit/real_space_span"
    }

    for soruce, targets in mapping.items():
        in_key1, in_key2 = soruce.split("/")
        value = task_hparams[in_key1][in_key2]

        # Set the value at all targets
        targets = [targets] if isinstance(targets, str) else targets
        for target in targets:
            out_key1, out_key2 = target.split("/")
            shared_hparams.set_value(out_key1, out_key2, value, overwrite=True)

    # Add all data folders
    data_folders = ("train_data", "val_data", "test_data", "aug_data")
    for df in data_folders:
        yaml = task_hparams.get_group(df)
        shared_hparams.add_group(yaml_string=yaml)

    # Save the updates parameters to a new location
    shared_hparams.save_current(os.path.join(out_dir, "train_hparams.yaml"))

    # Add weights to folder
    weights_folder = os.path.join(out_dir, "model")
    create_folders(weights_folder)
    out_weights_name = os.path.split(weights)[1]
    out_weights_path = os.path.join(weights_folder, out_weights_name)
    func = copy if copy_weights else os.symlink
    if copy_weights:
        print("Copying weights...")
    else:
        print("Symlinking weights...")
    func(weights, out_weights_path)

    # Add views (check for existence for future compatibility with 3D models)
    if os.path.exists(views_file):
        func(views_file, os.path.join(out_dir, "views.npz"))


def entry_func(args=None):
    parser = get_parser()
    args = vars(parser.parse_args(args))

    # Get arguments
    project_folder = os.path.abspath(args["project_folder"])
    copy_weights = args["copy_weights"]
    weights = os.path.abspath(args["weights_file"] or
                              get_best_model(project_folder + "/model"))
    out_dir = os.path.join(project_folder, args["out_dir"])

    # Get main hyperparamter file and check if correct modelt ype
    hparams = YAMLHParams(project_folder + "/train_hparams.yaml", no_log=True)
    tasks = hparams.get("tasks", False)
    if not tasks:
        print("[ERROR] Project must be of type 'MultiTask'.")
        sys.exit(0)

    # Branch out each sub-task
    create_folders(out_dir)
    for name, hparams_file in zip(tasks["task_names"], tasks["hparam_files"]):
        print("\n[*] Branching task %s" % name)
        # Reload the hparams in each iteration as we overwrite fields each time
        hparams = YAMLHParams(project_folder + "/train_hparams.yaml", no_log=False)
        # Get task specific parameters
        task_hparams = YAMLHParams(project_folder + "/%s" % hparams_file)
        branch(task_name=name,
               out_dir=os.path.join(out_dir, name),
               task_hparams=task_hparams,
               task_hparams_file=hparams_file,
               shared_hparams=hparams,
               weights=weights,
               copy_weights=copy_weights,
               views_file=os.path.join(project_folder, "views.npz"))



from .test_all import run_tests, run_all_tests


def get_test_packages():
    import pkgutil
    mods = list(filter(lambda x: x.ispkg, pkgutil.iter_modules(__path__)))
    return [m.name for m in mods]



import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import tensorflow as tf
import psutil
import numpy as np
import os
from tensorflow.keras.callbacks import Callback
from datetime import datetime
from mpunet.logging import ScreenLogger
from mpunet.utils.plotting import (imshow_with_label_overlay, imshow,
                                   plot_all_training_curves)


class DividerLine(Callback):
    """
    Simply prints a line to screen after each epoch
    """
    def __init__(self, logger=None):
        """
        Args:
            logger: An instance of a MultiPlanar Logger that prints to screen
                    and/or file
        """
        super().__init__()
        self.logger = logger or ScreenLogger()

    def on_epoch_end(self, epoch, logs=None):
        self.logger("-"*45 + "\n")


class LearningCurve(Callback):
    """
    On epoch end this callback looks for all csv files matching the 'csv_regex'
    regex within the dir 'out_dir' and attempts to create a learning curve for
    each file that will be saved to 'out_dir'.

    Note: Failure to plot a learning curve based on a given csv file will
          is handled in the plot_all_training_curves function and will not
          cause the LearningCurve callback to raise an exception.
    """
    def __init__(self, log_dir="logs", out_dir="logs", fname="curve.png",
                 csv_regex="*training.csv", logger=None, **plot_kwargs):
        """
        Args:
            log_dir: Relative path from the
            out_dir:
            fname:
            csv_regex:
            logger:
        """
        super().__init__()
        out_dir = os.path.abspath(out_dir)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        self.csv_regex = os.path.join(os.path.abspath(log_dir), csv_regex)
        self.save_path = os.path.join(out_dir, fname)
        self.logger = logger or ScreenLogger()
        self.plot_kwargs = plot_kwargs

    def on_epoch_end(self, epoch, logs={}):
        plot_all_training_curves(self.csv_regex,
                                 self.save_path,
                                 logy=True,
                                 raise_error=False,
                                 logger=self.logger,
                                 **self.plot_kwargs)


class MemoryConsumption(Callback):
    def __init__(self, max_gib=None, round_=2, logger=None):
        self.max_gib = max_gib
        self.logger = logger
        self.round_ = round_

    def on_epoch_end(self, epoch, logs={}):
        process = psutil.Process(os.getpid())
        mem_bytes = process.memory_info().rss
        mem_gib = round(mem_bytes / (1024**3), self.round_)
        logs['memory_usage_gib'] = mem_gib
        if self.max_gib and mem_gib >= self.max_gib:
            self.warn("Stopping training from callback 'MemoryConsumption'! "
                      "Total memory consumption of {} GiB exceeds limitation"
                      " (self.max_gib = {}) ".format(mem_gib, self.max_gib))
            self.model.stop_training = True


class DelayedCallback(object):
    """
    Callback wrapper that delays the functionality of another callback by N
    number of epochs.
    """
    def __init__(self, callback, start_from=0, logger=None):
        """
        Args:
            callback:   A tf.keras callback
            start_from: Delay the activity of 'callback' until this epoch
                        'start_from'
            logger:     An instance of a MultiPlanar Logger that prints to screen
                        and/or file
        """
        self.logger = logger or ScreenLogger()
        self.callback = callback
        self.start_from = start_from

    def __getattr__(self, item):
        return getattr(self.callback, item)

    def on_epoch_end(self, epoch, logs=None):
        if epoch >= self.start_from-1:
            self.callback.on_epoch_end(epoch, logs=logs)
        else:
            self.logger("[%s] Not active at epoch %i - will be at %i" %
                        (self.callback.__class__.__name__,
                         epoch+1, self.start_from))


class TrainTimer(Callback):
    """
    Appends train timing information to the log.
    If called prior to tf.keras.callbacks.CSVLogger this information will
    be written to disk.
    """
    def __init__(self, logger=None, max_minutes=None, verbose=1):
        super().__init__()
        self.logger = logger or ScreenLogger()
        self.max_minutes = int(max_minutes) if max_minutes else None
        self.verbose = bool(verbose)

        # Timing attributes
        self.train_begin_time = None
        self.prev_epoch_time = None

    def on_train_begin(self, logs=None):
        self.train_begin_time = datetime.now()

    def on_epoch_begin(self, epoch, logs=None):
        self.prev_epoch_time = datetime.now()

    def on_epoch_end(self, epoch, logs=None):
        # Compute epoch execution time
        end_time = datetime.now()
        epoch_time = end_time - self.prev_epoch_time
        train_time = end_time - self.train_begin_time

        # Update attributes
        self.prev_epoch_time = end_time

        # Add to logs
        train_hours = round(train_time.total_seconds() / 3600, 4)
        epoch_minutes = round(epoch_time.total_seconds() / 60, 4)
        logs["epoch_minutes"] = epoch_minutes
        logs["train_hours"] = train_hours

        if self.verbose:
            self.logger("[TrainTimer] Epoch time: %.2f minutes "
                        "- Total train time: %.2f hours"
                        % (epoch_minutes, train_hours))
        if self.max_minutes and train_hours*60 > self.max_minutes:
            self.logger("Stopping training. Training ran for {} minutes, "
                        "max_minutes of {} was specified on the TrainTimer "
                        "callback.".format(train_hours*60, self.max_minutes))
            self.model.stop_training = True


class FGBatchBalancer(Callback):
    """
    mpunet callback.

    Sets the forced FG fraction in a batch at each epoch to 1-recall over the
    validation data at the previous epoch
    """
    def __init__(self, train_data, val_data=None, logger=None):
        """
        Args:
            train_data: A mpunet.sequence object representing the
                        training data
            val_data:   A mpunet.sequence object representing the
                        validation data
            logger:     An instance of a MultiPlanar Logger that prints to screen
                        and/or file
        """
        super().__init__()
        self.data = (("train", train_data), ("val", val_data))
        self.logger = logger or ScreenLogger()
        self.active = True

    def on_epoch_end(self, epoch, logs=None):
        if not self.active:
            return None

        recall = logs.get("val_recall")
        if recall is None:
            self.logger("[FGBatchBalancer] No val_recall in logs. "
                        "Disabling callback. "
                        "Did you put this callback before the validation "
                        "callback?")
            self.active = False
        else:
            # Always at least 1 image slice
            fraction = max(0.01, 1 - recall)
            for name, data in self.data:
                if data is not None:
                    data.fg_batch_fraction = fraction
                    self.logger("[FGBatchBalancer] Setting FG fraction for %s "
                                "to: %.4f - Now %s/%s" % (name,
                                                          fraction,
                                                          data.n_fg_slices,
                                                          data.batch_size))


class MeanReduceLogArrays(Callback):
    """
    On epoch end, goes through the log and replaces any array entries with
    their mean value.
    """
    def __init__(self):
        super().__init__()

    def on_epoch_end(self, epoch, logs={}):
        for key, value in logs.items():
            if isinstance(value, (np.ndarray, list)):
                logs[key] = np.mean(value)


class PrintLayerWeights(Callback):
    """
    Print the weights of a specified layer every some epoch or batch.
    """
    def __init__(self, layer, every=10, first=10, per_epoch=False, logger=None):
        """
        Args:
            layer:      A tf.keras layer
            every:      Print the weights every 'every' batch or epoch if
                        per_epoch=True
            first:      Print the first 'first' elements of each weight matrix
            per_epoch:  Print after 'every' epoch instead of batch
            logger:     An instance of a MultiPlanar Logger that prints to screen
                        and/or file
        """
        super().__init__()
        if isinstance(layer, int):
            self.layer = self.model.layers[layer]
        else:
            self.layer = layer
        self.first = first
        self.every = every
        self.logger = logger or ScreenLogger()

        self.per_epoch = per_epoch
        if per_epoch:
            # Apply on every epoch instead of per batches
            self.on_epoch_begin = self.on_batch_begin
            self.on_batch_begin = lambda *args, **kwargs: None
        self.log()

    def log(self):
        self.logger("PrintLayerWeights Callback")
        self.logger("Layer:      ", self.layer)
        self.logger("Every:      ", self.every)
        self.logger("First:      ", self.first)
        self.logger("Per epoch:  ", self.per_epoch)

    def on_batch_begin(self, batch, logs=None):
        if batch % self.every:
            return
        weights = self.layer.get_weights()
        self.logger("Weights for layer '%s'" % self.layer)
        self.logger("Weights:\n%s" % weights[0].ravel()[:self.first])
        try:
            self.logger("Baises:\n%s" % weights[1].ravel()[:self.first])
        except IndexError:
            pass


class SaveOutputAs2DImage(Callback):
    """
    Save random 2D slices from the output of a given layer during training.
    """
    def __init__(self, layer, sequence, model, out_dir, every=10, logger=None):
        """
        Args:
            layer:    A tf.keras layer
            sequence: A MultiPlanar.sequence object from which batches are
                      sampled and pushed through the graph to output of layer
            model:    A tf.keras model object
            out_dir:  Path to directory (existing or non-existing) in which
                      images will be stored
            every:    Perform this operation every 'every' batches
        """
        super().__init__()
        self.every = every
        self.seq = sequence
        self.layer = layer
        self.epoch = None
        self.model = model
        self.logger = logger or ScreenLogger()

        self.out_dir = out_dir
        if not os.path.exists(out_dir):
            os.makedirs(self.out_dir)
        self.log()

    def log(self):
        self.logger("Save Output as 2D Image Callback")
        self.logger("Layer:      ", self.layer)
        self.logger("Every:      ", self.every)

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch = epoch

    def on_batch_end(self, batch, logs=None):
        if batch % self.every:
            return

        # Get output of layer
        self.model.predict_on_batch()
        sess = tf.keras.backend.get_session()
        X, _, _ = self.seq[0]
        outs = sess.run([self.layer.output], feed_dict={self.model.input: X})[0]
        if isinstance(outs, list):
            outs = outs[0]

        for i, (model_in, layer_out) in enumerate(zip(X, outs)):
            fig = plt.figure(figsize=(12, 6))
            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)

            # Plot model input and layer outputs on each ax
            chl1, axis, slice = imshow(ax1, model_in)
            chl2, _, _ = imshow(ax2, layer_out, axis=axis, slice=slice)

            # Set labels and save figure
            ax1.set_title("Model input - Channel %i - Axis %i - Slice %i"
                          % (chl1, axis,slice), size=22)
            ax2.set_title("Layer output - Channel %i - Axis %i - Slice %i"
                          % (chl2, axis, slice), size=22)

            fig.tight_layout()
            fig.savefig(os.path.join(self.out_dir, "epoch_%i_batch_%i_im_%i" %
                                     (self.epoch, batch, i)))
            plt.close(fig)


class SavePredictionImages(Callback):
    """
    Save images after each epoch of training of the model on a batch of
    training and a batch of validation data sampled from sequence objects.

    Saves the input image with ground truth overlay as well as the predicted
    label masks.
    """
    def __init__(self, train_data, val_data, outdir='images'):
        """
        Args:
            train_data: A mpunet.sequence object from which training
                        data can be sampled via the __getitem__ method.
            val_data:   A mpunet.sequence object from which validation
                        data can be sampled via the __getitem__ method.
            outdir:     Path to directory (existing or non-existing) in which
                        images will be stored.
        """
        super().__init__()

        self.train_data = train_data
        self.val_data = val_data
        self.save_path = os.path.abspath(os.path.join(outdir, "pred_images_at_epoch"))

        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

    def pred_and_save(self, data, subdir):
        # Get a random batch
        X, y, _ = data[np.random.randint(len(data), dtype=np.int64)]

        # Predict on the batch
        pred = self.model.predict(X)

        subdir = os.path.join(self.save_path, subdir)
        if not os.path.exists(subdir):
            os.mkdir(subdir)

        # Plot each sample in the batch
        for i, (im, lab, p) in enumerate(zip(X, y, pred)):
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(12, 6))
            lab = lab.reshape(im.shape[:-1] + (lab.shape[-1],))
            p = p.reshape(im.shape[:-1] + (p.shape[-1],))
            # Imshow ground truth on ax2
            # This function will determine which channel, axis and slice to
            # show and return so that we can use them for the other 2 axes
            chnl, axis, slice = imshow_with_label_overlay(ax2, im, lab, lab_alpha=1.0)

            # Imshow pred on ax3
            imshow_with_label_overlay(ax3, im, p, lab_alpha=1.0,
                                      channel=chnl, axis=axis, slice=slice)

            # Imshow raw image on ax1
            # Chose the same slice, channel and axis as above
            im = im[..., chnl]
            im = np.moveaxis(im, axis, 0)
            if slice is not None:
                # Only for 3D imges
                im = im[slice]
            ax1.imshow(im, cmap="gray")

            # Set labels
            ax1.set_title("Image", size=18)
            ax2.set_title("True labels", size=18)
            ax3.set_title("Prediction", size=18)

            fig.tight_layout()
            with np.testing.suppress_warnings() as sup:
                sup.filter(UserWarning)
                fig.savefig(os.path.join(subdir, str(i) + ".png"))
            plt.close(fig.number)

    def on_epoch_end(self, epoch, logs={}):
        self.pred_and_save(self.train_data, "train_%s" % epoch)
        if self.val_data is not None:
            self.pred_and_save(self.val_data, "val_%s" % epoch)



import numpy as np
import numpy.linalg as npl


def mgrid_to_points(mgrid):
    """
    Takes a NxD1xD2xD3 meshgrid or tuple(meshgrid) and outputs a D1*D2*D3xN
    matrix of coordinate points
    """
    points = np.empty(shape=(np.prod(mgrid[0].shape), len(mgrid)),
                      dtype=mgrid[0].dtype)
    for i in range(len(mgrid)):
        points[:, i] = mgrid[i].ravel()
    return points


def points_to_mgrid(points, grid_shape):

    mgrid = np.empty(shape=((points.shape[1],) + tuple(grid_shape)),
                     dtype=points.dtype)
    for i in range(points.shape[1]):
        mgrid[i] = points[:, i].reshape(grid_shape)

    return mgrid


def get_angle(v1, v2):
    v1_u = v1 / npl.norm(v1)
    v2_u = v2 / npl.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)))


def get_rotation_matrix(axis, angle_deg=None, angle_rad=None):
    """
    Modified from:
    https://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector

    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta degrees.
    """
    theta = angle_rad or np.deg2rad(angle_deg)
    axis = np.asarray(axis).ravel()
    axis = axis/np.linalg.norm(axis)
    a = np.cos(theta/2.0)
    b, c, d = -axis*np.sin(theta/2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d

    return np.array([[aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac)],
                     [2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab)],
                     [2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc]])


def _rotate_grid(grid, rot_mat):
    """
    Rotates a grid around its center by 'angle_deg' degrees counter clockwise
    around the vector 'axis'.
    """
    points = mgrid_to_points(grid)

    # Center, rotate and bring back the grid
    center = np.mean(points, axis=0).reshape((1, 3))
    c_points = points - center
    rotated = np.dot(rot_mat, c_points.T).T.astype(np.float32) + center

    return rotated



import unittest
import pkgutil


class TestImports(unittest.TestCase):

    needed_mods = ("mpunet", "matplotlib", "scipy", "numpy",
                   "nibabel", "pandas", "ruamel.yaml", "sklearn", "h5py",
                   "tensorflow", "psutil")

    def test_modules_exist(self):
        """ Check if needed modules exist (are visible to Python) """
        for mod in self.needed_mods:
            self.assertIsNot(pkgutil.find_loader(mod), None,
                             msg="Could not find loader for needed package "
                                 "'{}' - This could indicate that the package "
                                 "is not installed or is not visible to the "
                                 "current Python interpreter "
                                 "session.".format(mod))



from multiprocessing import current_process
from tensorflow.keras.utils import Sequence
from abc import ABC, abstractmethod
import numpy as np


class BaseSequence(Sequence, ABC):
    def __init__(self):
        super().__init__()

        # A dictionary mapping process names to whether the process has been
        # seeded
        self.is_seeded = {}

    def seed(self):
        # If multiprocessing the processes will inherit the RNG state of the
        # main process - here we reseed each process once so that the batches
        # are randomly generated across processes
        pname = current_process().name
        if pname not in self.is_seeded or not self.is_seeded[pname]:
            # Re-seed this process
            # If threading this will just re-seed MainProcess
            np.random.seed()
            self.is_seeded[pname] = True

    @abstractmethod
    def __len__(self):
        raise NotImplemented

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplemented

    def __call__(self):
        import tensorflow as tf
        def tensor_iter():
            """ Iterates the dataset, converting numpy arrays to tensors """
            for x, y, w in self:
                yield (tf.convert_to_tensor(x),
                       tf.convert_to_tensor(y),
                       tf.convert_to_tensor(w))
        return tensor_iter()

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]



from mpunet.image.queue import (LoadingPool, LimitationQueue,
                                LazyQueue, EagerQueue)


def validate_queue_type(queue_type, dataset, max_loaded, logger):
    if queue_type is LimitationQueue and (max_loaded is None or
                                          len(dataset) <= max_loaded):
        logger.warn("Falling back to 'Eager' queue for dataset {} due "
                    "to 'max_loaded' value of {} which is either 'None' or "
                    "larger than or equal to the total number of images ({}) "
                    "in the dataset.".format(dataset, max_loaded, len(dataset)))
        return EagerQueue
    return queue_type


def get_data_queues(train_dataset,
                    val_dataset,
                    train_queue_type,
                    val_queue_type,
                    max_loaded,
                    num_access_before_reload,
                    logger):
    """
    TODO.

    Returns:

    """
    map_ = {'eager': EagerQueue,
            'lazy': LazyQueue,
            'limitation': LimitationQueue}
    train_queue = validate_queue_type(map_[train_queue_type.lower()],
                                      train_dataset, max_loaded, logger)
    if val_queue_type:
        val_queue = validate_queue_type(map_[val_queue_type.lower()],
                                        val_dataset, max_loaded, logger)
    else:
        val_queue = None

    # If validation queue, get a loader pool object, shared across datasets
    if train_queue is LimitationQueue or val_queue is LimitationQueue:
        loading_pool = LoadingPool(n_threads=3,
                                   max_queue_size=max_loaded or None,
                                   logger=logger)
    else:
        loading_pool = None

    train_queue = train_queue(
            dataset=train_dataset,
            max_loaded=max_loaded,
            num_access_before_reload=num_access_before_reload,  # TODO
            preload_now=True,
            await_preload=True,
            loading_pool=loading_pool,
            logger=logger
    )
    if val_dataset:
        val_queue = val_queue(
            dataset=val_dataset,
            max_loaded=max_loaded,
            num_access_before_reload=num_access_before_reload,  # TODO
            preload_now=True,
            await_preload=False,
            loading_pool=loading_pool,
            logger=logger
        )
    else:
        val_queue = None
    return train_queue, val_queue



from mpunet.logging.default_logger import ScreenLogger
from threading import Lock, Thread
from queue import Queue
from time import sleep


def _load_func(load_queue, results_queue, load_errors_queue, lock, logger):
    """

    Args:
        load_queue:

    Returns:

    """
    while True:
        to_load, dataset_id = load_queue.get()
        try:
            to_load.load()
        except Exception as e:
            with lock:
                logger.warn("[ERROR in LoadingPool] "
                            "Could not load image '{}': {}".format(to_load, e))
            load_errors_queue.put((to_load, dataset_id))
        finally:
            results_queue.put((to_load, dataset_id))
            load_queue.task_done()


def _gather_loaded(output_queue, registered_datasets):
    while True:
        # Wait for studies in the output queue
        image_pair, dataset_id = output_queue.get(block=True)
        load_put_function = registered_datasets[dataset_id][0]
        load_put_function(image_pair)
        output_queue.task_done()


def _gather_errors(load_errors_queue, registered_datasets):
    while True:
        # Wait for studies in the output queue
        image_pair, dataset_id = load_errors_queue.get(block=True)
        error_put_function = registered_datasets[dataset_id][1]
        error_put_function(image_pair)
        load_errors_queue.task_done()


class LoadingPool:
    """
    Implements a multithreading SleepStudyBase loading queue
    """
    def __init__(self,
                 n_threads=5,
                 max_queue_size=50,
                 logger=None):
        """
        Initialize a StudyLoader object from a list of SleepStudyDataset objects

        Args:
            TODO
        """
        # Setup load thread pool
        self.logger = logger or ScreenLogger()
        self._load_queue = Queue(maxsize=max_queue_size)
        self._output_queue = Queue(maxsize=max_queue_size)
        self._load_errors_queue = Queue(maxsize=3)  # We probably want to raise
                                                    # an error if this queue
                                                    # gets to more than ~3!
        self.thread_lock = Lock()

        args = (self._load_queue, self._output_queue, self._load_errors_queue,
                self.thread_lock, self.logger)
        self.pool = []
        for _ in range(n_threads):
            p = Thread(target=_load_func, args=args, daemon=True)
            p.start()
            self.pool.append(p)

        # Prepare gathering thread
        self._registered_datasets = {}
        self.gather_loaded_thread = Thread(target=_gather_loaded,
                                           args=(self._output_queue,
                                                 self._registered_datasets),
                                           daemon=True)
        self.gather_errors_thread = Thread(target=_gather_errors,
                                           args=(self._load_errors_queue,
                                                 self._registered_datasets),
                                           daemon=True)
        self.gather_loaded_thread.start()
        self.gather_errors_thread.start()

    @property
    def qsize(self):
        """ Returns the qsize of the load queue """
        return self._load_queue.qsize

    @property
    def maxsize(self):
        """ Returns the maxsize of the load queue """
        return self._load_queue.maxsize

    def join(self):
        """ Join on all queues """
        self._load_queue.join()
        self._output_queue.join()
        self._load_errors_queue.join()

    def add_image_to_load_queue(self, image_pair, dataset_id):
        if dataset_id not in self._registered_datasets:
            raise RuntimeError("Dataset {} is not registered. "
                               "Call StudyLoader.register_dataset before adding"
                               " items from that dataset to the loading "
                               "queue".format(dataset_id))
        if self.qsize() == self.maxsize:
            self.logger.warn("Loading queue seems about to block! "
                             "(max_size={}, current={}). "
                             "Sleeping until loading queue is empty "
                             "again.".format(self.maxsize,
                                             self.qsize()))
            while self.qsize() > 1:
                sleep(1)
        self._load_queue.put((image_pair, dataset_id))

    def register_dataset(self, dataset_id, load_put_function, error_put_function):
        with self.thread_lock:
            if dataset_id in self._registered_datasets:
                raise RuntimeWarning("A dataset of ID {} has already been "
                                     "registered.".format(dataset_id))
            self._registered_datasets[dataset_id] = (
                load_put_function, error_put_function
            )

    def de_register_dataset(self, dataset_id):
        with self.thread_lock:
            del self._registered_datasets[dataset_id]



from tensorflow.keras import regularizers
from tensorflow.keras.layers import Input, Conv2D, BatchNormalization, \
                                    MaxPooling2D
import tensorflow as tf

from mpunet.models import UNet


def check_all_same_length(attributes, kwargs, target_length):
    for attr in attributes:
        v = kwargs[attr]
        if not isinstance(v, (tuple, list)):
            v = [v]
            kwargs[attr] = v
        if len(v) != target_length:
            raise ValueError("Must pass a '%s' of length %i (one for "
                             "each task) - got %s" % (attr, target_length, v))


class MultiTaskUNet2D(UNet):
    def __init__(self, task_names, **kwargs):
        """
        """
        if not isinstance(task_names, (tuple, list)):
            raise ValueError("'task_names' must be a list or tuple object.")
        self.n_tasks = len(task_names)
        self.task_IDs = tuple(task_names)
        self._n_classes = None

        # Check that each task received the right number of parameters
        check = ("n_classes", "dim", "out_activation", "n_channels")
        check_all_same_length(check, kwargs, target_length=self.n_tasks)

        # Store encoder layers for shared use
        self.encoder_layers = None
        self.out_layers = None

        # Init base UNet class
        super().__init__(**kwargs)

    def _init_encoder(self, init_filters, kernel_reg=None, name="encoder"):
        self.encoder_layers = {}
        filters = init_filters
        for i in range(self.depth):
            l_name = name + "_L%s" % i
            conv1 = Conv2D(int(filters * self.cf), self.kernel_size,
                           activation=self.activation, padding=self.padding,
                           kernel_regularizer=kernel_reg,
                           name=l_name + "_conv1")
            conv2 = Conv2D(int(filters * self.cf), self.kernel_size,
                           activation=self.activation, padding=self.padding,
                           kernel_regularizer=kernel_reg,
                           name=l_name + "_conv2")
            bn = BatchNormalization(name=l_name + "_BN")
            max_pool = MaxPooling2D(pool_size=(2, 2), name=l_name + "_pool")

            # Add to dict for potential reuse
            layers = {
                "layer%s/conv1" % i: conv1,
                "layer%s/conv2" % i: conv2,
                "layer%s/batch_norm" % i: bn,
                "layer%s/max_pool" % i: max_pool
            }
            self.encoder_layers.update(layers)

            # Update filter count and add bn layer to list for residual conn.
            filters *= 2
        return filters

    def _apply_encoder(self, task_input):
        residual_connections = []
        in_ = task_input
        for i in range(self.depth):
            conv1 = self.encoder_layers["layer%s/conv1" % i](in_)
            conv2 = self.encoder_layers["layer%s/conv2" % i](conv1)
            bn = self.encoder_layers["layer%s/batch_norm" % i](conv2)
            in_ = self.encoder_layers["layer%s/max_pool" % i](bn)
            residual_connections.append(bn)

        return in_, residual_connections

    def init_model(self):
        """
        Build the UNet model with the specified input image shape.
        """
        self.img_shape = tuple([t for t in zip(*self.img_shape)])
        inputs = [Input(shape=s,
                        name="Input_%s" % t) for s, t in zip(self.img_shape,
                                                             self.task_IDs)]

        # Apply regularization if not None or 0
        kr = regularizers.l2(self.l2_reg) if self.l2_reg else None

        """
        Encoding path
        """
        # Init the encoder layers
        filters = self._init_encoder(init_filters=64, kernel_reg=kr)

        out_layers = []
        outputs = []
        zipped = zip(self.task_IDs, inputs, self.n_classes, self.out_activation)
        for task, in_, n_classes, activation in zipped:
            with tf.name_scope("Task_%s" % task):
                with tf.name_scope("encoder"):
                    # Apply the encoder to all inputs
                    in_, res = self._apply_encoder(in_)

                # Bottom (no max-pool)
                with tf.name_scope("bottom"):
                    bn = self._create_bottom(in_, filters, kr, name=task)

                # Up-sampling
                with tf.name_scope("decoder"):
                    bn = self._create_upsample(bn, res, filters, kr, name=task)

                """
                Output modeling layer
                """
                with tf.name_scope("classifier"):
                    out_layer = Conv2D(n_classes, 1,
                                       activation=activation,
                                       name="%s" % task)
                    out = out_layer(bn)
                    out_layers.append(out_layer)
                    outputs.append(out)

        self.out_layers = tuple(out_layers)
        return inputs, outputs

    def log(self):
        self.logger("Multi-Task UNet Model Summary\n"
                    "-----------------------------")
        self.logger("N classes:         %s" % list(self.n_classes))
        self.logger("CF factor:         %.3f" % self.cf**2)
        self.logger("Depth:             %i" % self.depth)
        self.logger("l2 reg:            %s" % self.l2_reg)
        self.logger("Padding:           %s" % self.padding)
        self.logger("Conv activation:   %s" % self.activation)
        self.logger("Out activation:    %s" % list(self.out_activation))
        self.logger("Receptive field:   %s" % self.receptive_field)
        self.logger("N params:          %i" % self.count_params())
        self.logger("N tasks:           %i" % self.n_tasks)
        if self.n_tasks > 1:
            inputs = self.input
            outputs = self.output
        else:
            inputs = [self.input]
            outputs = [self.output]
        for i, (id_, in_, out) in enumerate(zip(self.task_IDs, inputs, outputs)):
            self.logger("\n--- Task %s ---" % id_)
            self.logger("In shape:  %s" % in_.shape)
            self.logger("Out shape: %s\n" % out.shape)



from mpunet.logging import ScreenLogger


def warn_sparse_param(logger):
    logger = logger or ScreenLogger
    sparse_err = "mpunet 0.1.3 or higher requires integer targets" \
                 " as opposed to one-hot encoded targets. Setting the 'sparse'" \
                 " parameter no longer has any effect and may not be allowed" \
                 " in future versions."
    logger.warn(sparse_err)



from mpunet.image.queue.base_queue import BaseQueue
from contextlib import contextmanager


class LazyQueue(BaseQueue):
    """
    Implements a queue-like object (same API interface as LoadQueue), but one
    that only loads data just-in-time when requested.
    This is useful for wrapping e.g. validation data in an object that behaves
    similar to the training queue object, but without consuming memory before
    needing to do validation.
    """
    def __init__(self, dataset, logger=None, **kwargs):
        """
        TODO
        Args:
            dataset:
            logger:
        """
        super(LazyQueue, self).__init__(
            dataset=dataset,
            logger=logger
        )
        self.logger("'Lazy' queue created:\n"
                    "  Dataset:      {}".format(self.dataset))
        self.logger("Images will be loaded just-in-time and unloaded when "
                    "not in use.".format(len(dataset)))

    @contextmanager
    def get_random_image(self):
        image = super().get_random_image()
        with image.loaded_in_context():
            yield image

    @contextmanager
    def get_image_by_idx(self, image_idx):
        image = super().get_image_by_idx(image_idx)
        with image.loaded_in_context():
            yield image

    @contextmanager
    def get_image_by_id(self, image_id):
        image = super().get_image_by_id(image_id)
        with image.loaded_in_context():
            yield image



from .funcs import init_callback_objects, remove_validation_callbacks
from .mcp_clean import ModelCheckPointClean
from .callbacks import (SavePredictionImages, PrintLayerWeights,
                        FGBatchBalancer, DividerLine,
                        SaveOutputAs2DImage, TrainTimer, LearningCurve,
                        MemoryConsumption, MeanReduceLogArrays)
from .validation import Validation, ValDiceScores



import numpy as np
from mpunet.logging.default_logger import ScreenLogger


class BaseQueue:
    """
    The base queue object defines the Queue API and stores basic attributes
    used across all queue objects
    The BaseQueue should normally not be initialized directly.
    """
    def __init__(self, dataset, logger=None):
        """
        TODO
        Args:
            datasets:
            logger:
        """
        self.dataset = dataset
        self.logger = logger or ScreenLogger()

    @property
    def all_loaded(self):
        raise NotImplemented

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        for i in range(len(self.dataset.images)):
            yield self.get_image_by_idx(i)

    def __getitem__(self, idx):
        return self.get_image_by_idx(idx)

    def get_image_iterator(self, max_load=None):
        load_inds = np.arange(len(self))
        if max_load and max_load < len(self):
            load_inds = np.random.choice(load_inds, max_load, False)
        for idx in load_inds:
            yield self.get_image_by_idx(idx)

    def get_random_image(self):
        return np.random.choice(self.dataset.images, 1)[0]

    def get_image_by_idx(self, image_idx):
        return self.dataset.images[image_idx]

    def get_image_by_id(self, image_id):
        return self.dataset.id_to_image[image_id]



from threading import Thread, Event
from contextlib import contextmanager
from queue import Queue
import numpy as np
import time


def _start(queue, stop_event):
    """
    Thread target function. Starts the ImageQueue._populate function which runs
    indefinitely until stop_event is set.

    Args:
        queue:      A reference to the ImageQueue object onto which the threads
                    apply.
        stop_event: An even that can be set in the main thread to stop
                    population of the ImageQueue
    """
    while not stop_event.is_set():
        queue._populate()


class ImageQueue(object):
    """
    Queue object handling loading ImagePair data from disk, preprocessing those
    images and adding them to a queue from which mpunet.sequence objects
    can pull them for training

    The queue will store a maximum number of loaded ImagePairs at a given time.
    When the queue is not full, one or more threads will perform the following
    sequence of actions in parallel to populate the queue:

    1) Select a random image from the ImagePairLoader object
    2) An entry function is called on the ImagePair that will perform some
       preprocessing operations
    3) The image is added to the queue

    Whenever 1) is performed, a sample may with a probability managed by
    attribute 'self.load_new_prob' instead re-add an image already loaded,
    processed and added to the queue. This process is much faster, and will be
    performed with a higher probability if the queue is being emptied too fast
    for the population to keep up.

    ImagePairs are pulled from the queue using the context manager method 'get':
    ...
    with image_queue.get() as image:
        # Do something with 'image', an ImagePair object
        f(image)
    ...
    When the with statement exits, one of two things happens in ImageQueue.get:
        1) If the ImagePair no longer exists in the queue, the ImageQueue
           invokes an exit function on the ImagePair. This function will
           usually free memory.
        2) If the ImagePair is still in queue - which may happen when the same
           image is re-added to queue to prevent queue exhaustion - the exit
           function is NOT invoked.

    TODO: Race conditions may occur in updating/referencing the dictionary
          storing the number of times images are currently loaded in queue,
          during executing of entry and exit functions and during calculation
          of load_new_prob.
          Update to a more robust queuing approach.
          Move to a TensorFlow queue based system perhaps?
    """
    def __init__(self, max_queue_size, image_pair_loader, entry_func=None,
                 entry_func_kw=None, exit_func=None, exit_func_kw=None):
        """
        Args:
            max_queue_size:    Int, the maximum number of ImagePair objects
                               to store in the queue at a given time
            image_pair_loader: The ImagePairLoader object from which images are
                               fetched.
            entry_func:        String giving name of method to call on the
                               ImagePair object at queue entry time.
            entry_func_kw:     Dict, keyword arguments to supply to entry_func
            exit_func:         String giving name of method to call on the
                               ImagePair object at queue exit time.
            exit_func_kw:      Dict, keyword arguments to supply to exit_func
        """
        # Reference Queue and ImagePairLoader objects
        self.queue = Queue(maxsize=max_queue_size)
        self.image_pair_loader = image_pair_loader

        # Initialize probability of loading a new (not currently in queue)
        # image to 1.0 (queue is empty at first anyway)
        self.load_new_prob = 1.0

        # Call the entry func when an image is added to the queue and the exit
        # func when the image leaves the queue
        self.entry_func = (entry_func, entry_func_kw or {})
        self.exit_func = (exit_func, exit_func_kw or {})

        # Store reference to all running threads
        self.threads = []

        # Store the number of times each image identifier is currently in the queue
        self.items_in_queue = 0
        self._last = 0
        self.no_new_counter = 0

        # Reference to images not in queue and IDs in queue
        self.num_times_in_queue = {image: 0 for image in self.image_pair_loader}

    @property
    def load_new_prob(self):
        return self._load_new_prob

    @load_new_prob.setter
    def load_new_prob(self, value):
        self._load_new_prob = np.clip(value, 0.05, 1.0)

    def set_entry_func(self, func_str, func_kw=None):
        self.entry_func = (func_str, func_kw or {})

    def set_exit_func(self, func_str, func_kw=None):
        self.exit_func = (func_str, func_kw or {})

    def wait_N(self, N):
        """
        Sleep until N images has been added to the queue

        Args:
            N: Int, number of images to wait for
        """
        cur = self.items_in_queue
        while self.items_in_queue < cur + N-1:
            time.sleep(1)

    @contextmanager
    def get(self):
        """
        Context manager method pulling an image from the queue and yielding it
        At yield return time the exit_func is called upon the image unless it
        has another reference later in the queue

        yields:
            an ImagePair from the queue
        """
        if self.items_in_queue < 0.1 * self.queue.maxsize:
            # If queue is almost empty, halt the main thread a bit
            self.wait_N(N=3)

        # Get the image from the queue
        image = self.queue.get()

        # Check if too high new_prob
        if self._last:
            diff = self.items_in_queue - self._last
            if diff > 0 or self.items_in_queue >= self.queue.maxsize-1:
                # If queue is increasing in size, increase load new prob
                self.load_new_prob *= 1.05
            elif diff < 0:
                # If queue is decreasing in size, decrease load new prob
                self.load_new_prob *= 0.95
        else:
            self._last = self.items_in_queue

        # Yield back
        yield image

        # Update reference attributes
        self.items_in_queue -= 1
        self.num_times_in_queue[image] -= 1

        # Call exit function on the object
        if self.num_times_in_queue[image] == 0:
            # Unload if last in the queue
            getattr(image, self.exit_func[0])(**self.exit_func[1])
            image.load_state = None

    def start(self, n_threads=3):
        """
        Start populating the queue in n_threads

        Args:
            n_threads: Number of threads to spin up
        """
        for _ in range(n_threads):
            stop_event = Event()
            thread = Thread(target=_start, args=(self, stop_event))
            thread.start()
            self.threads.append((thread, stop_event))

    def stop(self):
        """
        Stop populating the queue by invoking the stop event on all threads and
        wait for them to terminate.
        """
        print("Stopping %i threads" % len(self.threads))
        for _, event in self.threads:
            # Make sure no threads keep working after next addition to the Q
            event.set()
        for i, (t, _) in enumerate(self.threads):
            # Wait for the threads to stop
            print("   %i/%i" % (i+1, len(self.threads)), end="\r", flush=True)
            t.join()
        print("")

    @property
    def unique_in_queue(self):
        """
        Returns:
            Int, the current number of unique images in the queue
        """
        return sum([bool(m) for m in self.num_times_in_queue.values()])

    def await_full(self):
        """
        Halt main thread until queue object is populated to its max capacity
        """
        while self.items_in_queue < self.queue.maxsize:
            print("   Data queue being populated %i/%i" % (self.items_in_queue,
                                                           self.queue.maxsize),
                  end='\r', flush=True)
            time.sleep(1)

    def _populate(self):
        """
        Puts a random image into the queue. The ImagePair is either taken from
        the ImagePairLoader in an un-loaded state or from the already loaded,
        processed images stored in the current queue.

        This method should be continuously invoked from one of more threads
        to maintain a populated queue.
        """
        # With load_new_prob probability we chose not to reload a new image
        load_new = np.random.rand() < self.load_new_prob or \
                   (self.unique_in_queue < 0.2 * self.queue.maxsize)

        # Pick random image
        found = False
        while not found:
            image = self.image_pair_loader.images[np.random.randint(len(self.image_pair_loader))]
            already_loaded = bool(self.num_times_in_queue[image])
            found = load_new != already_loaded

        # Increment the image counter
        self.num_times_in_queue[image] += 1

        # If the image is not currently loaded, invoke the entry function
        if getattr(image, "load_state", None) != self.entry_func[0]:
            # Set load_state so that future calls dont try to load and
            # preprocess again
            image.load_state = self.entry_func[0]

            # Call entry function
            getattr(image, self.entry_func[0])(**self.entry_func[1])

        # Add it to the queue, block indefinitely until spot is free
        self.queue.put(image, block=True, timeout=None)

        # Increment in-queue counter
        self.items_in_queue += 1



from . import PatchSequence3D
from mpunet.interpolation.linalg import mgrid_to_points
import numpy as np


def standardize_strides(strides):
    if isinstance(strides, list):
        return tuple(strides)
    elif isinstance(strides, tuple):
        return strides
    else:
        return 3 * (int(strides),)


class SlidingPatchSequence3D(PatchSequence3D):
    def __init__(self, strides, no_log=False, *args, **kwargs):
        """
        strides: tuple (s1, s2, s3) of strides or integer s --> (s, s, s)
        """
        super().__init__(no_log=True, *args, **kwargs)

        # Stride attribute gives the number pixel distance between
        # patch samples in 3 dimensions
        self.strides = standardize_strides(strides)
        self.corners = self.get_patch_corners().astype(np.uint16)
        self.ind = np.arange(self.corners.shape[0])

        if not self.is_validation and not no_log:
            self.log()

    @property
    def n_samples(self):
        return self.corners.shape[0] * len(self.data)

    def get_patch_corners(self):
        xc = np.linspace(0, self.dim_r[0], self.strides[0]).astype(np.int)
        yc = np.linspace(0, self.dim_r[1], self.strides[1]).astype(np.int)
        zc = np.linspace(0, self.dim_r[2], self.strides[2]).astype(np.int)

        return mgrid_to_points(np.meshgrid(xc, yc, zc))

    def get_box_coords(self):
        return self.corners[np.random.choice(self.ind)]

    def get_base_patches(self, image_id):
        # Get sliding windows for X
        X = self.data[image_id][0]
        for xc, yc, zc in self.corners:
            patch = X[xc:xc + self.dim, yc:yc + self.dim, zc:zc + self.dim]
            yield patch, (xc, yc, zc)

    def log(self):
        self.logger("Sequence Generator: %s" % self.__class__.__name__)
        self.logger("Box dimensions:     %s" % self.dim)
        self.logger("Image dimensions:   %s" % self.im_dim)
        self.logger("Sample space dim:   %s" % self.dim_r)
        self.logger("Strides:            %s" % list(self.strides))
        self.logger("N boxes:            %s" % self.corners.shape[0])
        self.logger("Batch size:         %s" % self.batch_size)
        self.logger("N fg slices/batch:  %s" % self.n_fg_slices)



"""
mpunet train script
Optimizes a mpunet model in a specified project folder

Typical usage:
--------------
mp init_project --name my_project --data_dir ~/path/to/data
cd my_project
mp train --num_GPUs=1
--------------
"""


from argparse import ArgumentParser
import os


def get_argparser():
    parser = ArgumentParser(description='Fit a mpunet model defined '
                                        'in a project folder. '
                                        'Invoke "init_project" to start a '
                                        'new project.')
    parser.add_argument("--project_dir", type=str, default='./',
                        help="Path to a mpunet project directory. "
                             "Defaults to the current directory.")
    parser.add_argument("--num_GPUs", type=int, default=1,
                        help="Number of GPUs to use for this job (default=1)")
    parser.add_argument("--force_GPU", type=str, default="",
                        help="Manually set the CUDA_VISIBLE_DEVICES env "
                             "variable to this value "
                             "(force a specific set of GPUs to be used)")
    parser.add_argument("--continue_training", action="store_true",
                        help="Continue the last training session")
    parser.add_argument("--overwrite", action='store_true',
                        help='Overwrite previous training session at the '
                             'project path')
    parser.add_argument("--just_one", action="store_true",
                        help="For testing purposes, run only on the first "
                             "training and validation samples.")
    parser.add_argument("--no_val", action="store_true",
                        help="Do not perform validation (must be set if no "
                             "validation set exists)")
    parser.add_argument("--no_images", action="store_true",
                        help="Do not save sample images during training")
    parser.add_argument("--debug", action="store_true",
                        help="Set tfbg CLI wrapper on the session object")
    parser.add_argument("--wait_for", type=str, default="",
                        help="Wait for PID to terminate before starting the "
                             "training process.")
    parser.add_argument("--train_images_per_epoch", type=int, default=2500,
                        help="Number of training images to sample in each "
                             "epoch")
    parser.add_argument("--val_images_per_epoch", type=int, default=3500,
                        help="Number of training images to sample in each "
                             "epoch")
    parser.add_argument("--max_loaded_images", type=int, default=None,
                        help="Set a maximum number of (training) images to "
                             "keep loaded in memory at a given time. Images "
                             "will be cycled every '--num_access slices.'. "
                             "Default=None (all loaded).")
    parser.add_argument("--epochs", type=int, default=None,
                        help="Overwrite the number of epochs as specified in "
                             "the hyperparameters file")
    parser.add_argument("--num_access", type=int, default=50,
                        help="Only effective with --max_loaded_images set. "
                             "Sets the number of times an images stored in "
                             "memory may be accessed (e.g. for sampling an "
                             "image slice) before it is replaced by another "
                             "image. Higher values makes the data loader "
                             "less likely to block. Lower values ensures that "
                             "images are sampled across all images of the "
                             "dataset. Default=50.")
    return parser


def validate_project_dir(project_dir):
    if not os.path.exists(project_dir) or \
            not os.path.exists(os.path.join(project_dir, "train_hparams.yaml")):
        raise RuntimeError("The script was launched from directory:\n'%s'"
                           "\n... but this is not a valid project folder.\n\n"
                           "* Make sure to launch the script from within a "
                           "MultiPlanarNet project directory\n"
                           "* Make sure that the directory contains a "
                           "'train_hparams.yaml' file." % project_dir)


def validate_args(args):
    """
    Checks that the passed commandline arguments are valid

    Args:
        args: argparse arguments
    """
    if args.continue_training and args.overwrite:
        raise ValueError("Cannot both continue training and overwrite the "
                         "previous training session. Remove the --overwrite "
                         "flag if trying to continue a previous training "
                         "session.")
    if args.train_images_per_epoch <= 0:
        raise ValueError("train_images_per_epoch must be a positive integer")
    if args.val_images_per_epoch <= 0:
        raise ValueError("val_images_per_epoch must be a positive integer. "
                         "Use --no_val instead.")
    if args.force_GPU and args.num_GPUs != 1:
        raise ValueError("Should not specify both --force_GPU and --num_GPUs")
    if args.num_GPUs < 0:
        raise ValueError("num_GPUs must be a positive integer")


def validate_hparams(hparams):
    """
    Limited number of checks performed on the validity of the hyperparameters.
    The file is generally considered to follow the semantics of the
    mpunet.bin.defaults hyperparameter files.

    Args:
        hparams: A YAMLHParams object
    """
    # Tests for valid hparams
    if hparams["fit"].get("class_weights") and hparams["fit"]["loss"] not in \
            ("SparseFocalLoss",):
        # Only currently supported losses
        raise ValueError("Invalid loss function '{}' used with the "
                         "'class_weights' "
                         "parameter".format(hparams["fit"]["loss"]))
    if hparams["fit"]["loss"] == "WeightedCrossEntropyWithLogits":
        if not bool(hparams["fit"]["class_weights"]):
            raise ValueError("Must specify 'class_weights' argument with loss"
                             "'WeightedCrossEntropyWithLogits'.")
        if not hparams["build"]["out_activation"] == "linear":
            raise ValueError("Must use out_activation: linear parameter with "
                             "loss 'WeightedCrossEntropyWithLogits'")
    if not hparams["train_data"]["base_dir"]:
        raise ValueError("No training data folder specified in parameter file.")


def remove_previous_session(project_folder):
    """
    Deletes various mpunet project folders and files from
    [project_folder].

    Args:
        project_folder: A path to a mpunet project folder
    """
    import shutil
    # Remove old files and directories of logs, images etc if existing
    paths = [os.path.join(project_folder, p) for p in ("images",
                                                       "logs",
                                                       "tensorboard",
                                                       "views.npz",
                                                       "views.png")]
    for p in filter(os.path.exists, paths):
        if os.path.isdir(p):
            shutil.rmtree(p)
        else:
            os.remove(p)


def get_logger(project_dir, overwrite_existing):
    """
    Initialises and returns a Logger object for a given project directory.
    If a logfile already exists at the specified location, it will be
    overwritten if continue_training == True, otherwise raises RuntimeError

    Args:
        project_dir: Path to a mpunet project folder
        overwrite_existing: Whether to overwrite existing logfile in project_dir

    Returns:
        A mpunet Logger object initialized in project_dir
    """
    # Define Logger object
    from mpunet.logging import Logger
    try:
        logger = Logger(base_path=project_dir,
                        print_to_screen=True,
                        overwrite_existing=overwrite_existing)
    except OSError as e:
        raise RuntimeError("[*] A training session at '%s' already exists."
                           "\n    Use the --overwrite flag to "
                           "overwrite." % project_dir) from e
    return logger


def get_gpu_monitor(num_GPUs, logger):
    """
    Args:
        num_GPUs: Number of GPUs to train on
        logger: A mpunet logger object that will be passed to
                the GPUMonitor

    Returns:
        If num_GPUs >= 0, returns a GPUMonitor object, otherwise returns None
    """
    if num_GPUs >= 0:
        # Initialize GPUMonitor in separate fork now before memory builds up
        from mpunet.utils.system import GPUMonitor
        gpu_mon = GPUMonitor()
    else:
        gpu_mon = None
    return gpu_mon


def set_gpu(gpu_mon, args):
    """
    Sets the GPU visibility based on the passed arguments. Takes an already
    initialized GPUMonitor object. Sets GPUs according to args.force_GPU, if
    specified, otherwise sets first args.num_GPUs free GPUs on the system.

    Stops the GPUMonitor process once GPUs have been set
    If gpu_mon is None, this function does nothing

    Args:
        gpu_mon: An initialized GPUMonitor object or None
        args: argparse arguments

    Returns: The number of GPUs that was actually set (different from
    args.num_GPUs if args.force_GPU is set to more than 1 GPU)
    """
    num_GPUs = args.num_GPUs
    if gpu_mon is not None:
        if not args.force_GPU:
            gpu_mon.await_and_set_free_GPU(N=num_GPUs, sleep_seconds=120)
        else:
            gpu_mon.set_GPUs = args.force_GPU
            num_GPUs = len(args.force_GPU.split(","))
        gpu_mon.stop()
    return num_GPUs


def get_data_sequences(project_dir, hparams, logger, args):
    """
    Loads training and validation data as specified in the hyperparameter file.
    Returns a batch sequencer object for each dataset, not the  ImagePairLoader
    dataset itself. The preprocessing function may make changes to the hparams
    dictionary.

    Args:
        project_dir: A path to a mpunet project
        hparams: A YAMLHParams object
        logger: A mpunet logging object
        args: argparse arguments

    Returns:
        train: A batch sequencer object for the training data
        val: A batch sequencer object for the validation data,
             or None if --no_val was specified
        hparams: The YAMLHParams object
    """
    from mpunet.preprocessing import get_preprocessing_func
    func = get_preprocessing_func(hparams["build"].get("model_class_name"))
    hparams['fit']['flatten_y'] = True
    hparams['fit']['max_loaded'] = args.max_loaded_images
    hparams['fit']['num_access'] = args.num_access
    train, val, hparams = func(hparams=hparams,
                               logger=logger,
                               just_one=args.just_one,
                               no_val=args.no_val,
                               continue_training=args.continue_training,
                               base_path=project_dir)
    return train, val, hparams


def get_model(project_dir, train_seq, hparams, logger, args):
    """
    Initializes a tf.keras Model from mpunet.models as specified in
    hparams['build']. If args.continue_training, the best previous model stored
    in [project_dir]/models will be loaded.

    If hparams["build"]["biased_output_layer"] is True, sets the bias weights
    on the final conv. layer so that a zero-input gives an output of class
    probabilities equal to the class frequencies of the training set.

    Args:
        project_dir: A path to a mpunet project folder
        train_seq: A mpunet.sequences object for the training data
        hparams: A mpunet YAMLHParams object
        logger: A mpunet logging object
        args: argparse arguments

    Returns:
        model: The model to fit
        org_model: The original, non-GPU-distributed model
                   (Same as model if num_GPUs==1)
    """
    from mpunet.models import model_initializer
    # Build new model (or continue training an existing one)
    hparams["build"]['flatten_output'] = True
    model = model_initializer(hparams=hparams,
                              continue_training=args.continue_training,
                              project_dir=project_dir,
                              logger=logger)
    # Initialize weights in final layer?
    if not args.continue_training and hparams["build"].get("biased_output_layer"):
        from mpunet.utils.utils import set_bias_weights_on_all_outputs
        set_bias_weights_on_all_outputs(model,
                                        train_seq.image_pair_queue,
                                        hparams,
                                        logger)
    return model


def save_final_weights(model, project_dir, logger=None):
    """
    Saves the weights of 'model' to [project_dir]/model/model_weights.h5

    Args:
        model: A tf.keras Model object
        project_dir: A path to a mpunet project
        logger: mpunet logging object, or None
    """
    if not os.path.exists("%s/model" % project_dir):
        os.mkdir("%s/model" % project_dir)
    model_path = "%s/model/model_weights.h5" % project_dir
    if logger:
        logger("Saving current model to: %s" % model_path)
    model.save_weights(model_path)


def run(project_dir, gpu_mon, logger, args):
    """
    Runs training of a model in a mpunet project directory.

    Args:
        project_dir: A path to a mpunet project
        gpu_mon: An initialized GPUMonitor object
        logger: A mpunet logging object
        args: argparse arguments
    """
    # Read in hyperparameters from YAML file
    from mpunet.hyperparameters import YAMLHParams
    hparams = YAMLHParams(project_dir + "/train_hparams.yaml", logger=logger)
    validate_hparams(hparams)

    # Wait for PID to terminate before continuing?
    if args.wait_for:
        from mpunet.utils import await_PIDs
        await_PIDs(args.wait_for)

    # Prepare sequence generators and potential model specific hparam changes
    train, val, hparams = get_data_sequences(project_dir=project_dir,
                                             hparams=hparams,
                                             logger=logger,
                                             args=args)

    # Set GPU visibility and create model with MirroredStrategy
    set_gpu(gpu_mon, args)
    import tensorflow as tf
    with tf.distribute.MirroredStrategy().scope():
        model = get_model(project_dir=project_dir, train_seq=train,
                          hparams=hparams, logger=logger, args=args)

        # Get trainer and compile model
        from mpunet.train import Trainer
        trainer = Trainer(model, logger=logger)
        trainer.compile_model(n_classes=hparams["build"].get("n_classes"),
                              reduction=tf.keras.losses.Reduction.NONE,
                              **hparams["fit"])

    # Debug mode?
    if args.debug:
        from tensorflow.python import debug as tfdbg
        from tensorflow.keras import backend as K
        K.set_session(tfdbg.LocalCLIDebugWrapperSession(K.get_session()))

    # Fit the model
    hparams["fit"]["n_epochs"] = args.epochs or hparams["fit"]["n_epochs"]
    try:
        _ = trainer.fit(train=train, val=val,
                        train_im_per_epoch=args.train_images_per_epoch,
                        val_im_per_epoch=args.val_images_per_epoch,
                        hparams=hparams, no_im=args.no_images, **hparams["fit"])
    except KeyboardInterrupt:
        gpu_mon.stop()
    finally:
        save_final_weights(model, project_dir, logger)


def entry_func(args=None):
    """
    Function called from mp to init training
    1) Parses command-line arguments
    2) Validation command-line arguments
    3) Checks and potentially deletes a preious version of the project folder
    4) Initializes a logger and a GPUMonitor object
    5) Calls run() to start training

    Args:
        args: None or arguments passed from mp
    """
    # Get and check args
    args = get_argparser().parse_args(args)
    validate_args(args)

    # Check for project dir
    project_dir = os.path.abspath(args.project_dir)
    validate_project_dir(project_dir)
    os.chdir(project_dir)

    # Get project folder and remove previous session if --overwrite
    if args.overwrite:
        remove_previous_session(project_dir)

    # Get logger object. Overwrites previous logfile if args.continue_training
    logger = get_logger(project_dir, args.continue_training)
    logger("Fitting model in path:\n%s" % project_dir)

    # Start GPU monitor process, if num_GPUs > 0
    gpu_mon = get_gpu_monitor(args.num_GPUs, logger)

    try:
        run(project_dir=project_dir, gpu_mon=gpu_mon, logger=logger, args=args)
    except Exception as e:
        if gpu_mon is not None:
            gpu_mon.stop()
        raise e


if __name__ == "__main__":
    entry_func()



from tensorflow.keras import callbacks as tfcb
from .callbacks import DelayedCallback


def init_callback_objects(callbacks, logger):
    """
    Initialize a list of tf.keras/custom callback descriptors.

    Args:
        callbacks: List of callback descriptions. Each list entry should be
                   either a dictionary of the format
                   {"class_name": <name_of_CB_class>,
                   "kwargs": {<dict_with_kwargs>}}
                   ... or an already initialized callback object
                   (which will be skipped).
        logger:    A mpunet logger object logging to screen and/or file

    Returns:
        A list of initialized callbacks
    """
    from mpunet import callbacks as tcb

    cb_objs = []
    cb_dict = {}
    for i, callback in enumerate(callbacks):
        if not isinstance(callback, dict):
            # CB already initialized
            cb = callback
            kwargs = {"params": "?"}
            cls_name = callback.__class__.__name__
            start_from = 0
        else:
            kwargs = callback["kwargs"]
            cls_name = callback["class_name"]
            start_from = callback.get("start_from")
            if callback.get("pass_logger"):
                kwargs["logger"] = logger
            try:
                cb = getattr(tfcb, cls_name, None) or getattr(tcb, cls_name)
            except AttributeError as e:
                raise ValueError("No callback named %s" % cls_name) from e
            cb = cb(**kwargs)
        if start_from:
            logger("OBS: '%s' activates at epoch %i" % (cls_name, start_from))
            cb = DelayedCallback(callback=cb,
                                 start_from=start_from,
                                 logger=logger)
        cb_objs.append(cb)
        cb_dict[cls_name] = cb
        logger("[%i] Using callback: %s(%s)" % (i+1, cb.__class__.__name__,
                                                ", ".join(["%s=%s" % (a, kwargs[a]) for a in kwargs])))

    return cb_objs, cb_dict


def remove_validation_callbacks(callbacks, logger=None):
    """
    Removes all callbacks that rely on validation data

    Takes a list of uninitialized callbacks data, enumerates them and removes
    each entry if one or more of its parameters in 'kwargs' mentions 'val'.

    Args:
        callbacks: A list of dictionaries, each representing a callback

    Returns:
        None, operates in-place
    """
    for i, callback in enumerate(callbacks):
        val_dependent_params = []
        for param in callback["kwargs"].values():
            val_dependent_params.append("val" in str(param).lower())
        if any(val_dependent_params):
            if logger:
                logger("Removing callback with parameters: {} "
                       "(needs validation data)".format(callback))
            callbacks.pop(i)



import sklearn.preprocessing as preprocessing
import numpy as np


def assert_scaler(scaler):
    return scaler in preprocessing.__dict__


def get_scaler(scaler, *args, ignore_less_eq=None, **kwargs):
    scaler = preprocessing.__dict__[scaler]
    return MultiChannelScaler(scaler=scaler,
                              ignore_less_eq=ignore_less_eq,
                              *args, **kwargs)


def apply_scaling(X, scaler, ignore_less_eq=None):
    # Get scaler
    multi_scaler = get_scaler(scaler, ignore_less_eq=ignore_less_eq)

    # Fit and apply transformation
    return multi_scaler.fit_transform(X)


class MultiChannelScaler(object):
    def __init__(self, scaler, *args, ignore_less_eq=None, **kwargs):
        # Store scaler class and passed parameters
        self.scaler_class = scaler
        self.scaler_args = args
        self.scaler_kwargs = kwargs
        self.ignore_less_eq = ignore_less_eq

        # Store list of initialized scalers fit to each channel
        self.scalers = []

        # Store number of channels
        self.n_channels = None

    def __str__(self):
        return "MultiChannelScaler(scaler_class='{}', ignore_less_eq={})".format(
            self.scaler_class.__name__,
            self.ignore_less_eq
        )

    def __repr__(self):
        return str(self)

    def fit(self, X, *args, **kwargs):
        if X.ndim != 4:
            raise ValueError("Invalid shape for X (%s)" % X.shape)

        # Set number of channels
        self.n_channels = X.shape[-1]

        if self.ignore_less_eq is not None:
            if not isinstance(self.ignore_less_eq, (list, tuple, np.ndarray)):
                self.ignore_less_eq = [self.ignore_less_eq] * self.n_channels
            if not len(self.ignore_less_eq) == self.n_channels:
                raise ValueError("'ignore_less_eq' should be a list of length "
                                 "'n_channels'. Got {} for n_channels={}".format(
                    self.ignore_less_eq, self.n_channels
                ))

        scalers = []
        for i in range(self.n_channels):
            sc = self.scaler_class(*self.scaler_args, **self.scaler_kwargs)
            xs = X[..., i]
            if self.ignore_less_eq is not None:
                xs = xs[np.where(xs > self.ignore_less_eq[i])]
            sc.fit(xs.reshape(-1, 1), *args, **kwargs)
            scalers.append(sc)

        self.scalers = scalers
        return self

    def transform(self, X, *args, **kwargs):
        if X.shape[-1] != self.n_channels:
            raise ValueError("Invalid input of dimension %i, expected "
                             "last axis with %i channels" % (X.ndim,
                                                             self.n_channels))

        # Prepare volume like X to store results
        transformed = np.empty_like(X)
        for i in range(self.n_channels):
            scl = self.scalers[i]
            s = scl.transform(X[..., i].reshape(-1, 1), *args, **kwargs)
            transformed[..., i] = s.reshape(X.shape[:-1])

        return transformed

    def fit_transform(self, X, *args, **kwargs):
        self.fit(X, *args, **kwargs)
        return self.transform(X, *args, **kwargs)
